from typing import Optional, Dict, List, Union
import os
import copy
import math
import time
import json

import numpy as np
import ROOT

from quickstats import DescriptiveEnum
from quickstats.maths.numerics import pretty_value
from quickstats.components.basics import WSArgument
from quickstats.components import AnalysisObject
from quickstats.utils.common_utils import parse_config

class FitMode(DescriptiveEnum):
    HYBRID  = (0, "Execute both conditional and unconditional fits for a given set of POIs")
    UNCOND  = (1, "Execute unconditional fit for a given set of POIs")
    COND    = (2, "Execute conditional fit with specified conditional values for a given set of POIs")
    NOMINAL = (3, "Execute fit without any POIs")
    NOFIT   = (4, "Do not perform any fit")

class Likelihood(AnalysisObject):
    
    def __init__(self, filename:str, poi_name:Optional[Union[str, List[str]]]=None,
                 data_name:str='combData', 
                 config:Optional[Dict]=None,
                 verbosity:Optional[Union[int, str]]="INFO"):
        config = parse_config(config)
        config['filename']  = filename
        config['poi_name']  = poi_name
        config['data_name'] = data_name
        config['verbosity'] = verbosity
        self._inherit_init(super().__init__, **config)
        self.roofit_result = None
    
    @staticmethod
    def _parse_poi_val(pois:"ROOT.RooArgSet", poi_val:Optional[Union[Dict[str, float], float]]=None):
        if poi_val is None:
            poi_val = {}
            for poi in pois:
                poi_val[poi.GetName()] = poi.getVal()
        elif not isinstance(poi_val, dict):
            poi_val_tmp = poi_val
            poi_val = {}
            for poi in pois:
                poi_val[poi.GetName()] = poi_val_tmp
        else:
            # fill in missing values
            for poi in pois:
                poi_name = poi.GetName()
                if poi_name not in poi_val:
                    poi_val[poi_name] = poi.getVal()
        return poi_val
    
    @staticmethod
    def get_combined_fit_result(uncond_fit_result:Dict, cond_fit_result:Dict):
        combined_fit_result = {}
        nll_uncond = uncond_fit_result['nll']
        nll_cond = cond_fit_result['nll']
        pnll = nll_cond - nll_uncond
        qmu  = 2 * pnll
        combined_fit_result['pnll'] = pnll
        combined_fit_result['qmu']  = qmu
        poi_val = cond_fit_result['mu']
        if (qmu >= 0):
            ndof = len(poi_val)
            if ndof == 1:
                # ndof = 1 case
                poi_name = list(poi_val)[0]
                x0 = poi_val[poi_name]
                significance = math.sqrt(qmu)
                pvalue = 1 - ROOT.Math.normal_cdf(significance, 1, x0)
            else:
                x0 = list(set(poi_val.values()))
                if len(x0) > 1:
                    combined_fit_result['significance'] = None
                    combined_fit_result['pvalue'] = None
                else:
                    pvalue = ROOT.Math.chisquared_cdf_c(qmu, ndof, x0[0])
                    significance = ROOT.RooStats.PValueToSignificance(pvalue)
            combined_fit_result['significance'] = significance
            combined_fit_result['pvalue'] = pvalue
        else:
            combined_fit_result['significance'] = None
            combined_fit_result['pvalue'] = None
        return combined_fit_result
        
    def nll_fit(self, poi_val:Optional[Union[Dict[str, float], float]]=None,
                mode:Union[int, str, FitMode]="uncond",
                snapshot_name:Optional[str]=AnalysisObject.kCurrentSnapshotName,
                do_minos:bool=False):
        """
        Perform a fit based on the Maximum Likelihood Estimate (MLE). In particular,
        the negative log-likelihood (NLL) is minimized.
        
        Parameters
        ----------------------------------------------------------------------------
        poi_val: (optional) float or dictionary of float
            Value(s) of parameter(s) of interest when performing a conditional MLE fit.
            If value is a float, then all parameters of interest will be assigned the given value.
            If value is a dictionary, it represents a map between the parameter name and the assigned value.
            Any missing parameter name will be assigned its current value (as in the snapshot).
        mode: int, str or FitMode, default = "uncond"
            Evaluation mode. Choose from:
                0 or "hybrid"  : Evaluate unconditional + conditional NLL
                1 or "uncond"  : Evaluate unconditional NLL (globally minimized NLL)
                2 or "cond"    : Evaluate conditional NLL (NLL minimized at fixed mu values)
                3 or "nominal" : Evaluate post-fit NLL (POI states remain as is)
                4 or "nofit"   : Evaluate pre-fit NLL (No fit performed)
        snapshot_name: (Optional) str or dictionary, default=AnalysisObject.kCurrentSnapshotName
            Name of snapshot to load before fitting. If value is a dictionary, it must be of the
            form {"cond": <conditional_fit_snapshot>, "uncond": <unconditional_fit_snapshot>} and
            can only be used if evaluation mode = "hybrid".
        do_minos: bool, default = False
            Evalate minos error for all parameters of interest.
        """
        mode = FitMode.parse(mode)
        if not isinstance(self.poi, ROOT.RooArgSet):
            pois = ROOT.RooArgSet(self.poi)
        else:
            pois = self.poi
            
        if snapshot_name is None:
            self.save_snapshot(self.kTempSnapshotName, WSArgument.MUTABLE)
            snapshot_name = self.kTempSnapshotName
        if isinstance(snapshot_name, dict) and (mode != FitMode.HYBRID):
            raise ValueError('snapshot name is allowed to be a dictionary only when hybrid evaluation mode is used')
        if isinstance(snapshot_name, str):
            snapshot_name = {
                "uncond": snapshot_name,
                "cond"  : snapshot_name
            }
        if "uncond" not in snapshot_name:
            snapshot_name["uncond"] = self.kCurrentSnapshotName
            self.stdout.info(f'INFO: No snapshot given for unconditional fit. The snapshot "{self.kCurrentSnapshotName}" '
                             'will be used by default')
        if "cond" not in snapshot_name:
            snapshot_name["cond"] = self.kCurrentSnapshotName
            self.stdout.info(f'INFO: No snapshot given for conditional fit. The snapshot "{self.kCurrentSnapshotName}" '
                             'will be used by default')
        if set(snapshot_name.keys()) != set(["uncond", "cond"]):
            raise ValueError('when snapshot name is given as a dictionary, it must be of the '
            'form {"cond": <conditional_fit_snapshot>, "uncond": <unconditional_fit_snapshot>}')
        
        self.load_snapshot(snapshot_name["cond"])
        poi_val = self._parse_poi_val(pois, poi_val)
        
        nll_uncond = None
        nll_cond   = None
        muhat      = None
        
        tmp_do_minos = self.minimizer.config['minos']
        
        self.minimizer.config['minos'] = do_minos
        if do_minos:
            self.minimizer.minos_set = pois
        
        start_time = time.time()
        roofit_result = {}
        muhat = {}
        muhat_errlo = {}
        muhat_errhi = {}
        
        # no poi fit
        if mode == FitMode.NOMINAL:
            # restore initial parameter values after each fit
            self.load_snapshot(snapshot_name["cond"])
            if len(pois) > 0:
                # fix other pois if they are not studied
                for poi in self.model.pois:
                    poi.setConstant(1)
                for poi in pois:
                    poi.setConstant(0)
            self.minimizer.create_nll()
            prefit_nll = self.minimizer.nll.getVal()
            fit_status = self.minimizer.minimize()
            roofit_result = self.minimizer.fit_result
            self.save_snapshot("nllFit", WSArgument.MUTABLE)
            postfit_nll = self.minimizer.nll.getVal()
            fit_time = time.time() - start_time
            
        # unconditional nll
        if mode in [FitMode.HYBRID, FitMode.UNCOND, FitMode.NOFIT]:
            # restore initial parameter values after each fit
            self.load_snapshot(snapshot_name["uncond"])
            # fix other pois if they are not studied
            for poi in self.model.pois:
                poi.setConstant(1)
            for poi in pois:
                poi.setConstant(0)
            if mode == FitMode.NOFIT:
                self.minimizer.create_nll()
                fit_status_uncond = None
                roofit_result['uncond'] = None
            else:
                fit_status_uncond = self.minimizer.minimize()
                roofit_result['uncond'] = self.minimizer.fit_result
            self.save_snapshot("uncondFit", WSArgument.MUTABLE)
            nll_uncond = self.minimizer.nll.getVal()
            for poi in pois:
                poi_name = poi.GetName()
                muhat[poi_name]       = poi.getVal()
                muhat_errlo[poi_name] = poi.getErrorLo()
                muhat_errhi[poi_name] = poi.getErrorHi()
        fit_time_uncond = time.time() - start_time
        # conditional nll
        if mode in [FitMode.HYBRID, FitMode.COND]:
            # restore initial parameter values after each fit
            self.load_snapshot(snapshot_name["cond"])
            # fix other pois if they are not studied
            for poi in self.model.pois:
                poi.setConstant(1)
            for poi in pois:
                val = poi_val[poi.GetName()]
                # profiled poi
                if val is None:
                    poi.setConstant(0)
                else:
                    poi.setConstant(1)
                    poi.setVal(val)
            fit_status_cond = self.minimizer.minimize()
            roofit_result['cond'] = self.minimizer.fit_result
            self.save_snapshot("condFit", WSArgument.MUTABLE)
            nll_cond = self.minimizer.nll.getVal()
        fit_time_cond = time.time() - fit_time_uncond - start_time
        
        fit_result = {}
        self.stdout.info("INFO: NLL evaluation completed with")
        if mode in [FitMode.HYBRID, FitMode.UNCOND, FitMode.NOFIT]:
            best_fit_str = ", ".join([f"{name} = {value:.5f}" for name, value in muhat.items()])
            self.stdout.info("best fit : ".rjust(15) + f"{best_fit_str}")
            self.stdout.info("uncond NLL = ".rjust(15) + f"{nll_uncond:.5f}")
            fit_result['uncond_fit'] = {
                'muhat': muhat,
                'muhat_errlo': muhat_errlo,
                'muhat_errhi': muhat_errhi,
                'nll': nll_uncond,
                'status': fit_status_uncond,
                'time': fit_time_uncond
            }
        if mode in [FitMode.HYBRID, FitMode.COND]:
            mu = {name:pretty_value(value) for name, value in poi_val.items()}
            mu_str = ", ".join([f"{name} = {pretty_value(value)}" for name, value in poi_val.items()])
            self.stdout.info("mu : ".rjust(15) + f"{mu_str}")
            # also add the best-fit value for profiled POIs
            for name, value in poi_val.items():
                if value == None:
                    muhat[name] = self.model.workspace.var(name).getVal()
            self.stdout.info("cond NLL = ".rjust(15)+f"{nll_cond:.5f}")
            fit_result['cond_fit'] = {
                'mu': mu,
                'muhat': muhat,
                'nll': nll_cond,
                'status': fit_status_cond,
                'time': fit_time_cond
            }
            
        if mode == FitMode.HYBRID:
            pnll = nll_cond - nll_uncond
            qmu  = 2*pnll
            self.stdout.info("PNLL = ".rjust(15) + "{:.5f}".format(nll_cond - nll_uncond))
            fit_result['pnll'] = pnll
            fit_result['qmu']  = qmu
            
            if (qmu >= 0):
                ndof = len(pois)
                if ndof == 1:
                    # ndof = 1 case
                    poi_name = pois.first().GetName()
                    x0 = poi_val[poi_name]
                    significance = math.sqrt(qmu)
                    pvalue = 1 - ROOT.Math.normal_cdf(significance, 1, x0)
                else:
                    x0 = list(set(poi_val.values()))
                    if len(x0) > 1:
                        fit_result['significance'] = None
                        fit_result['pvalue'] = None
                    else:
                        pvalue = ROOT.Math.chisquared_cdf_c(qmu, ndof, x0[0])
                        significance = ROOT.RooStats.PValueToSignificance(pvalue)
                fit_result['significance'] = significance
                fit_result['pvalue'] = pvalue
            else:
                fit_result['significance'] = None
                fit_result['pvalue'] = None
                
        if mode != FitMode.NOMINAL:
            self.stdout.info("time = ".rjust(15) + \
                             "(uncond_fit) {:.3f}, (cond_fit) {:.3f}".format(fit_time_uncond, fit_time_cond))
            fit_result['total_time'] = fit_time_uncond + fit_time_cond
        else:
            self.stdout.info("prefit NLL = ".rjust(18) + f"{prefit_nll:.5f}")
            self.stdout.info("postfit NLL = ".rjust(18) + f"{postfit_nll:.5f}")
            self.stdout.info("time = ".rjust(18) + f"{fit_time:.3f}")
            fit_result['prefit_nll'] = prefit_nll
            fit_result['postfit_nll'] = postfit_nll
            fit_result['total_time'] = fit_time
        
        self.roofit_result = roofit_result
        
        # finallizing
        self.minimizer.config['minos'] = tmp_do_minos
        self.minimizer.minos_set = ROOT.RooArgSet()
        
        return fit_result
        
    def evaluate_nll(self, poi_val:Optional[Union[Dict[str, float], float]]=None,
                     mode:Union[int, str, FitMode]="uncond",
                     snapshot_name:Optional[str]=AnalysisObject.kCurrentSnapshotName):
        """
        Evalute post-fit NLL value
        
        Parameters
        ------------------------------------------------------------------------------
        poi_val: (optional) float or dictionary of float
            Value(s) of parameter(s) of interest when performing a conditional MLE fit.
            If value is a float, then all parameters of interest will be assigned the given value.
            If value is a dictionary, it represents a map between the parameter name and the assigned value.
            Any missing parameter name will be assigned its current value (as in the snapshot).
        mode: int, str or FitMode, default = "uncond"
            Evaluation mode. Choose from:
                0: Evaluate nll_mu - nll_muhat
                1: Evaluate unconditional NLL (nll_muhat)
                2: Evaluate conditional NLL (nll_mu)
                3: Evaluate post-fit NLL (POI state remain as is)
        snapshot_name: (Optional) str, default=AnalysisObject.kCurrentSnapshotName
            Name of snapshot to load before fitting.
        """
        fit_result = self.nll_fit(poi_val=poi_val, mode=mode, snapshot_name=snapshot_name)
        mode = FitMode.parse(mode)
        if mode == FitMode.HYBRID:
            return fit_result['pnll']
        elif mode == FitMode.UNCOND:
            return fit_result['uncond_fit']['nll']
        elif mode == FitMode.COND:
            return fit_result['cond_fit']['nll']
        elif mode == FitMode.NOMINAL:
            return fit_result['postfit_nll']


from typing import Dict, List, Optional, Union, Tuple, Any, Callable
import os
import re
import sys
import copy
import fnmatch

from .enums import GeneralEnum
from .abstract_object import AbstractObject

from dataclasses import MISSING, _MISSING_TYPE


class BasicConfigTypes(GeneralEnum):
    NONE        = (0, lambda x: x is None)
    ANY         = (1, lambda x: True)
    STR         = (2, lambda x: isinstance(x, str))
    FLOAT       = (3, lambda x: isinstance(x, (int, float)))
    BOOL        = (4, lambda x: isinstance(x, bool))
    INT         = (5, lambda x: isinstance(x, int))
    LIST        = (6, lambda x: isinstance(x, list))
    DICT        = (7, lambda x: isinstance(x, dict))
    
    def __new__(cls, value:int, validate_method:Callable):
        obj = object.__new__(cls)
        obj._value_ = value
        obj.validate = validate_method
        return obj

class CompositeConfigTypes(object):
    
    def __init__(self, signature:str):
        nest_level = signature.count("[")
        if nest_level == 0:
            # interpret as basic type
            config_type = BasicConfigTypes.parse(signature)
            self.name = config_type.name
            self.validate = config_type.validate
        else:
            if signature.count("]") != nest_level:
                raise ValueError(f'Incomplete bracket in the type signature "{signature}"')
            type_strings = signature.strip("]").split("[")
            config_types = []
            for type_string in type_strings:
                try:
                    config_type = BasicConfigTypes.parse(type_string)
                except Exception as e:
                    raise ValueError(f'Failed to parse the type signature "{signature}". An error '
                                     f'was received when parsing the component "{type_string}": {e}')
                config_types.append(config_type)
            if len(config_types) != (nest_level + 1):
                raise ValueError(f'Found spurious brackets in the type signature "{signature}"')
            if not all(config_type in [BasicConfigTypes.LIST, BasicConfigTypes.DICT] \
                       for config_type in config_types[:-1]):
                raise ValueError(f'Invalid nested type signature "{signature}": only dictionary or list nestings are allowed')
            validate = None
            for config_type in config_types[::-1]:
                validate = self._create_nested_lambda(config_type, validate)
            self.name = signature
            self.validate = validate
            
    def _create_nested_lambda(self, config_type, lambda_func):
        if lambda_func is None:
            return config_type.validate
        if config_type == BasicConfigTypes.LIST:
            return lambda x : isinstance(x, list) and all(lambda_func(y) for y in x)
        elif config_type == BasicConfigTypes.DICT:
            return lambda x : isinstance(x, dict) and all(lambda_func(y) for y in x.values())
        return None

class ConfigComponent(object):
    
    @property
    def name(self):
        return self._name
    
    @name.setter
    def name(self, value):
        assert isinstance(value, str)
        self._name = value
    
    @property
    def dtypes(self):
        return self._dtypes
    
    @dtypes.setter
    def dtypes(self, value):
        if isinstance(value, (tuple, list)):
            self._dtypes = [CompositeConfigTypes(v) for v in value]
        else:
            self._dtypes = [CompositeConfigTypes(value)]
    
    @property
    def required(self):
        return self._required
    
    @required.setter
    def required(self, value):
        assert isinstance(value, bool) 
        self._required = value
        
    @property
    def required_strict(self):
        return self._required_strict
    
    @required_strict.setter
    def required_strict(self, value):
        assert isinstance(value, bool) 
        self._required_strict = value        
        
    @property
    def default(self):
        return self._default
    
    @default.setter
    def default(self, value):
        if value == MISSING:
            self._default = MISSING
        else:
            self.type_check(value)
            self._default = value
            
    @property
    def default_factory(self):
        return self._default_factory
    
    @default_factory.setter
    def default_factory(self, value):
        if value == MISSING:
            self._default_factory = MISSING
        else:
            self.type_check(value())
            self._default_factory = value            
    
    @property
    def description(self):
        return self._description
    
    @description.setter
    def description(self, value):
        assert isinstance(value, (str, type(None)))
        self._description = value
    
    @property
    def example(self):
        return self._example
    
    @example.setter
    def example(self, value):
        assert isinstance(value, (str, type(None)))
        self._example = value    
    
    @property
    def choice(self):
        return self._choice
    
    @choice.setter
    def choice(self, value):
        if value is None:
            self._choice = None
        else:
            assert isinstance(value, (tuple, list))
            for v in value:
                self.type_check(v)
            self._choice = value
        
    @property    
    def deprecated(self):
        return self._deprecated
    
    @deprecated.setter
    def deprecated(self, value):
        assert isinstance(value, bool)
        self._deprecated = value
        
    @property
    def deprecation_message(self):
        return self._deprecation_message   
    
    @deprecation_message.setter
    def deprecation_message(self, value):
        assert isinstance(value, (str, type(None)))
        self._deprecation_message = value
    
    def __init__(self, dtypes:Union[str, Tuple[str], List[str]]="ANY",
                 name:str="",
                 required:bool=False,
                 default:Optional[Any]=MISSING,
                 default_factory:Optional[Callable]=MISSING,
                 description:Optional[str]=None,
                 example:Optional[str]=None,
                 choice:Optional[List[Any]]=None,
                 deprecated:bool=False,
                 deprecation_message:Optional[str]=None):
        self.name = name
        self.dtypes = dtypes
        self.required = required
        self.required_strict = False
        self.choice = choice
        self.default = default
        self.default_factory = default_factory
        if ((not isinstance(self.default, _MISSING_TYPE)) and
            (not isinstance(self.default_factory, _MISSING_TYPE))):
            raise ValueError('cannot specify both default and default_factory')
        self.description = description
        self.example = example
        self.deprecated = deprecated
        self.deprecation_message = deprecation_message      
        
    def type_check(self, value):
        passed = any(dtype.validate(value) for dtype in self.dtypes)
        if not passed:
            raise RuntimeError(f'Type check failed for the config component "{self.name}". '
                               f'Allowed types: {", ".join([dtype.name for dtype in self.dtypes])}')
            
    def has_default(self):
        return ((not isinstance(self.default, _MISSING_TYPE)) or 
                (not isinstance(self.default_factory, _MISSING_TYPE)))
    
    def get_default(self):
        if not isinstance(self.default, _MISSING_TYPE):
            return self.default
        if not isinstance(self.default_factory, _MISSING_TYPE):
            return self.default_factory()
        raise RuntimeError('default value not set')

    def validate(self, value=MISSING):
        if isinstance(value, _MISSING_TYPE):
            # throw error if missing value but no default value available
            if self.required or (self.required_strict and (not self.has_default())):
                raise RuntimeError(f'Missing value for the config component "{self.name}":\n'
                                   f'{self.get_explain_text()}')
            return self.get_default()
        self.type_check(value)
        if (self.choice is not None) and (value not in self.choice):
            raise RuntimeError(f'Invalid value for the config component "{self.name}". '
                               f'Supported values: {self.choice}')
        return value
    
    def get_explain_text(self, leftmargin:int=0, linebreak:int=100):
        components = {
            "dtype"       : "|".join([dtype.name for dtype in self.dtypes]),
            "required"    : str(self.required),
        }
        if self.has_default():
            components["default"] = str(self.get_default())
        components["description"] = self.description
        if self.example is not None:
            components["example"] = self.example
        if self.deprecated:
            components["deprecation"] = self.deprecation_message
        if self.choice is not None:
            components["choice"] = ", ".join([f"{i}" for i in self.choice])
            
        from quickstats.utils.common_utils import itemize_dict
        text = itemize_dict(components, leftmargin=leftmargin, linebreak=linebreak)
        return text

    def explain(self, linebreak:int=100):
        sys.stdout.write(self.get_explain_text(linebreak=linebreak))
    
class ConfigParser(AbstractObject):
    
    @property
    def config_format(self):
        return self._config_format
    
    @config_format.setter
    def config_format(self, value):
        if not isinstance(value, dict):
            raise ValueError("config format must be a dictionary")
        copied_value = copy.deepcopy(value)
        component_map = self.parse_config_format(copied_value)
        self._config_format = copied_value
        self._component_map = component_map
        
    @property
    def component_map(self):
        return self._component_map        
        
    def __init__(self, config_format:Dict,
                 required_components:Optional[List[str]]=None,
                 optional_components:Optional[List[str]]=None,
                 disable_message:bool=False,
                 verbosity:Optional[Union[int, str]]="INFO"):
        self.config_format = config_format
        self.set_components_requirement(required_components, optional_components)
        self.disable_message = disable_message
        
    def set_components_requirement(self, required_components:Optional[List[str]]=None,
                                   optional_components:Optional[List[str]]=None):
        
        # require everything by default
        if required_components is None:
            required_components = ["*"]
        if optional_components is None:
            optional_components = []            
        for label, component in self.component_map.items():
            component.required_strict = False
            for required_component in required_components:
                if fnmatch.fnmatch(label, required_component):
                    component.required_strict = True
                    break
            for optional_component in optional_components:
                if fnmatch.fnmatch(label, required_component):
                    component.required_strict = False
                    break
                    
    def get_required_components(self):
        components = [component for component in self.component_map.values() if component.required_strict]
        return components                
    
    def get_config_component(self, component_label:str):
        if component_label in self.component_map:
            return self.component_map[component_label]
        for label, component in self.component_map.items():
            label = label.replace("?:", "*").replace("?", "*")
            if fnmatch.fnmatch(component_label, label):
                return component
        return None
        
    def get_explain_text(self, node_label:Optional[str]="",
                         required_only:bool=False,
                         indent:str="    ", bullet:str="=>",
                         linebreak:int=100):
        if required_only:
            required_labels = set()
            for component in self.get_required_components():
                tokens = component.name.split(":")
                required_labels |= set([":".join(tokens[:i+1]) for i in range(len(tokens))])
            required_labels = list(required_labels)
        else:
            required_labels = None
        def traverse_node(config_node:Dict, level:int=0,
                          current_text:str="", config_label:str="",
                          write:bool=False, filter_labels:Optional[List[str]]=None):
            margin = indent * level
            indent_text = margin + bullet
            for key, value in config_node.items():
                if isinstance(value, ConfigComponent):
                    if (filter_labels is not None) and (value.name not in filter_labels):
                        continue
                    current_text += write * (indent_text + key + "\n")
                    current_text += write * (value.get_explain_text(leftmargin=len(indent_text.expandtabs()),
                                                                    linebreak=linebreak))
                elif isinstance(value, dict):
                    new_config_label = self._get_config_label(key, config_label)
                    new_write = write
                    if (not write):
                        match_label = new_config_label.replace("?:", "*").replace("?", "*")                        
                        if fnmatch.fnmatch(node_label, match_label):
                            new_write = True
                    if (filter_labels is not None) and (new_config_label not in filter_labels):
                        continue                            
                    current_text += new_write * (indent_text + key + "\n")
                    current_text = traverse_node(value, level + int(new_write),
                                                 current_text, new_config_label,
                                                 new_write, filter_labels)
                else:
                    raise RuntimeError("invalid config node")
            return current_text
        
        if self.get_config_component(node_label) is not None:
            component = self.get_config_component(node_label)
            text = bullet + node_label + "\n"
            text += component.get_explain_text(linebreak=linebreak)
        else:
            text = traverse_node(self.config_format, write=node_label=="",
                                 filter_labels=required_labels)
        if not text:
            if required_only:
                self.stdout.critical('ERROR: No required components matches the requested scope. '
                                     'Try using "required_only=False" instead.')
            elif node_label:
                self.stdout.critical(f'ERROR: Config node "{node_label}" does not exist.')
            else:
                self.stdout.critical(f'ERROR: Config format is empty.')                
        return text
    
    def explain(self, node_label:Optional[str]="",
                required_only:bool=False,
                indent:str="    ",
                bullet:str="=>",
                linebreak:int=100):
        sys.stdout.write(self.get_explain_text(node_label=node_label,
                                               required_only=required_only,
                                               indent=indent,
                                               bullet=bullet,
                                               linebreak=linebreak))
        
    def parse_config_format(self, config_format, label:str="",
                            component_map:Optional[Dict]=None):
        if component_map is None:
            component_map = {}
        if isinstance(config_format, ConfigComponent):
            config_format.name = label
            if label in component_map:
                raise RuntimeError(f"duplicated config component: {label}")
            component_map[label] = config_format                
        elif isinstance(config_format, dict):
            for key, value in config_format.items():
                new_label = self._get_config_label(key, label)
                if not isinstance(key, str):
                    raise ValueError(f'found non-string key in the config component "{new_label}"')
                self.parse_config_format(value, new_label, component_map)
        else:
            raise ValueError(f'found unsupported format in the config component "{label}"')
        return component_map
    
    def parse_config(self, config:Union[Dict, str]):
        if isinstance(config, str):
            extension = os.path.splitext(config)[-1]
            if not self.disable_message:
                self.stdout.info(f'INFO: Reading configuration file "{config}".')
            if extension == ".json":
                import json
                with open(config, "r") as file:
                    config = json.load(file)
            elif extension == ".yaml":
                import yaml
                with open(config, "r") as file:
                    config = yaml.safe_load(file)
            else:
                raise RuntimeError(f'Unsupported config format "{extension.strip(".")}. '
                                   'Allowed formats are: json, yaml')
        if not isinstance(config, dict):
            raise ValueError("config must be a dictionary or path to a json/yaml file")
        resolved_config = copy.deepcopy(config)
        self.validate_and_resolve(resolved_config, self.config_format)
        return resolved_config
    
    def _get_config_label(self, key:str, prefix:str=""):
        return f"{prefix}:{key}" if prefix else key
        
    def validate_and_resolve(self, config, formatter, prefix:str="",
                             extra_configs:Optional[List]=None):
        if extra_configs is None:
            extra_configs = []       
        if not isinstance(config, dict):
            return None
        if not isinstance(formatter, dict):
            raise RuntimeError("invalid config formatter")
        config_keys = list(config.keys())
        formatter_keys = list(formatter.keys())
        # optional layer
        if (formatter_keys == ["?"]) and isinstance(formatter["?"], dict):
            if (("*" not in formatter["?"]) and set(config_keys).issubset(set(formatter["?"].keys()))) or \
               (("*" in formatter["?"]) and set(formatter["?"].keys()).issubset(set(config_keys))):
                return self.validate_and_resolve(config, formatter["?"], prefix, extra_configs)
            else:
                formatter_keys = ["*" if k == "?" else k for k in formatter_keys]
        extra_keys = list(set(config_keys) - set(formatter_keys))
        if "*" not in formatter_keys:
            extra_configs += [self._get_config_label(key, prefix) for key in extra_keys]
        for key in formatter_keys:
            config_label = self._get_config_label(key, prefix)
            if (key == "*") and (key not in formatter):
                component = formatter["?"]
            else:
                component = formatter[key]
            if isinstance(component, ConfigComponent):
                if key in config:
                    config[key] = component.validate(config[key])
                elif key == "*":
                    for config_key in extra_keys:
                        config[config_key] = component.validate(config[config_key])
                else:
                    # only fill default if component is strictly required
                    if component.required_strict:
                        config[key] = component.validate(MISSING)
                        if not self.disable_message:
                            self.stdout.info(f'INFO: Auto-filled missing config component "{config_label}" with '
                                         f'default value "{config[key]}".')
            else:
                if key == "*":
                    config_keys = extra_keys
                else:
                    if key not in config:
                        config[key] = {}
                    config_keys = [key]
                for config_key in config_keys:
                    self.validate_and_resolve(config[config_key], component,
                                              config_label, extra_configs)
        return extra_configs


from typing import Optional, List, Dict

from .rooproc_base_action import RooProcBaseAction

class RooProcHelperAction(RooProcBaseAction):
    
    def _execute(self, processor:"quickstats.RooProcessor", **params):
        return processor
    
    def execute(self, processor:"quickstats.RooProcessor", global_vars:Optional[Dict]=None):
        params = self.get_formatted_parameters(global_vars)
        processor = self._execute(processor, **params)
        self.executed = True
        return processor


from typing import Union, List, Optional, Dict
from itertools import repeat
import os
import sys
import copy

import numpy as np
import pandas as pd

from quickstats import semistaticmethod, AbstractObject
from quickstats.utils.common_utils import execute_multi_tasks
from quickstats.utils.data_conversion import downcast_dataframe
from quickstats.analysis import AnalysisBase

class NTupleConversionTool(AnalysisBase):
    
    REQUIRED_CONFIG_COMPONENTS = ["paths:*"]
    
    DEFAULT_COLUMNS = {}
    
    DEFAULT_NTUPLE_SAMPLE_BASENAME = "{sample_name}.root"
    DEFAULT_ARRAY_SAMPLE_BASENAME  = "{sample_name}.{fmt}"
    
    DEFAULT_H5_KEY = "analysis_data"
    
    def __init__(self, analysis_config:Optional[Union[Dict, str]]=None,
                 ntuple_dir:Optional[str]=None,
                 array_dir:Optional[str]=None,
                 verbosity:Optional[Union[int, str]]="INFO",
                 **kwargs):
        super().__init__(analysis_config=analysis_config,
                         ntuple_dir=ntuple_dir,
                         array_dir=array_dir,
                         disable_config_message=True,
                         verbosity=verbosity,
                         **kwargs)
        self.stdout.info(f'INFO: Initialized ntuple directory as "{self.get_directory("ntuple")}".')
        self.stdout.info(f'INFO: Initialized array directory as "{self.get_directory("array")}".')
        self.stdout.info(f'INFO: Initialized ntuple treename as "{self.treename}".')
        
    @semistaticmethod
    def _save(self, df, outpath:str, fmt:str="csv",
              mode:str="a", complevel:int=None):
        
        assert (fmt in ["csv", "h5"])
        assert (mode in ["w", "a", "r"])
        
        if fmt == "csv":
            if mode in ["w", "r"]:
                header = True
            else:
                header = False
            df.to_csv(outpath, mode=mode, index=False, header=header)
        elif fmt == "h5":
            df.to_hdf(outpath, key=self.DEFAULT_H5_KEY, mode=mode,
                      index=False, complevel=complevel)
                
    @semistaticmethod
    def _postprocess(self, df, **kwargs):
        return df
    
    @semistaticmethod
    def _finalize(self, df, **kwargs):
        pass

    @semistaticmethod
    def _convert(self, filename:str, sample_name:str, treename:str,
                 outdir:str, columns:Optional[Union[List, Dict]]=None,
                 apply_columns:Optional[Dict]=None,
                 drop_columns:Optional[List]=None,
                 fmt:str="csv",
                 downcast:bool=True,
                 complevel:int=None,
                 chunksize:int=100000,
                 library:str="quickstats",
                 kwargs:Optional[Dict]=None):
        
        if columns is None:
            columns = copy.deepcopy(self.DEFAULT_COLUMNS)

        self.stdout.info(f'INFO: Processing file "{filename}"')

        #NTupleProcessTool.remove_csv(sample, outdir)
        
        if library == "quickstats":
            # does not support chunksize
            from quickstats.utils.data_conversion import root2dataframe
            df = root2dataframe(filename, treename, columns=columns,
                                remove_non_standard_types=False,
                                downcast=downcast, library="root")
            df_iter = [df]
        elif library == "root_pandas":
            import root_pandas
            _columns = ["noexpand:" + k for k in columns.values()]
            df_iter = root_pandas.read_root(filename, 
                                            key=treename,
                                            columns=_columns,
                                            chunksize=chunksize)
        else:
            raise RuntimeError(f"invalid library: {library}")
                               
        for i, df in enumerate(df_iter):
            
            if len(df) == 0: 
                continue            
            
            if library == "root_pandas":
                df.columns = list(columns)
                df.reset_index(drop=True, inplace=True)
                
                # Fix some columns which are just arrays of length 1
                for col in df.columns:
                    if df[col].dtype == 'object' and df[col].iloc[0].size == 1: df[col] = df[col].str[0]
                        
                if downcast:
                    downcast_dataframe(df)
                
            if apply_columns is not None:
                for key, value in apply_columns.items():
                    result = df.apply(lambda x: value(x), axis=1)
                    df[list(key)] = df.apply(lambda x: value(x), axis=1)
            
            df = self._postprocess(df, **kwargs, sample_name=sample_name)
            
            if drop_columns is not None:
                df = df.drop(drop_columns, axis=1)
                
            outpath = self.get_array_sample_path(sample_name=sample_name, dirname=outdir, fmt=fmt)
            if i == 0:
                self.stdout.info(f'INFO: Saving data array as "{outpath}"')
                self._save(df, outpath=outpath, fmt=fmt, 
                           mode='w', complevel=complevel)
            else:
                self._save(df, outpath=outpath, fmt=fmt, 
                           mode='a', complevel=complevel)
                
            self._finalize(df, **kwargs, sample_name=sample_name, outdir=outdir,
                           fmt=fmt, complevel=complevel, iteration=i)
            
    @semistaticmethod
    def get_ntuple_sample_path(self, sample_name:Dict, dirname:str):
        basename = self.DEFAULT_NTUPLE_SAMPLE_BASENAME.format(sample_name=sample_name)
        return os.path.join(dirname, basename)
    
    @semistaticmethod
    def get_array_sample_path(self, sample_name:Dict, dirname:str, fmt:str):
        basename = self.DEFAULT_ARRAY_SAMPLE_BASENAME.format(sample_name=sample_name, fmt=fmt)
        return os.path.join(dirname, basename)    
                 
    def convert_samples(self, samples:Optional[List[str]]=None,
                        kinematic_regions:Optional[List]=None,
                        columns:Optional[Union[List, Dict]]=None,
                        apply_columns:Optional[Dict]=None,
                        drop_columns:Optional[List]=None,
                        library:str="quickstats",
                        fmt:Optional[str]=None,
                        downcast:Optional[bool]=None,
                        complevel:Optional[int]=None,
                        chunksize:int=100000,
                        parallel:int=0,
                        **kwargs):
        if fmt is None:
            fmt = self.get_analysis_data_format()
        if fmt not in ["csv", "h5"]:
            raise ValueError("only csv or h5 formats are allowed")
            
        if (fmt != "h5") and (complevel is not None):
            raise ValueError("compression level is only allowed in h5 format")
        
        if downcast is None:
            if fmt == "csv":
                downcast = False
            else:
                downcast = True
                
        if self.all_samples:
            samples = self.resolve_samples(samples)
        if samples is None:
            raise RuntimeError("no samples specified")
            
        if kinematic_regions is None:
            kinematic_regions = self.all_kinematic_regions
        # no kinematic regions defined, use inclusive data with no sub-directories
        if not kinematic_regions:
            kinematic_regions = [""]
            
        if library == "quickstats":
            import ROOT
            if ROOT.IsImplicitMTEnabled():
                ROOT.DisableImplicitMT()
            if parallel != 0:
                parallel = 0
                sys.stdout.write("INFO: Disabling multi-processing when using ROOT backend\n")            
        
        ntuple_dir = self.get_directory("ntuple")
        array_dir = self.get_directory("array")
        
        for kinematic_region in kinematic_regions:
            if kinematic_region:
                self.stdout.info(f'INFO: Converting ntuples in the kinematic region "{kinematic_region}"')
                ntuple_subdir = os.path.join(ntuple_dir, kinematic_region)
                array_subdir = os.path.join(array_dir, kinematic_region)
            else:
                ntuple_subdir = ntuple_dir
                array_subdir  = array_dir
                
            if not os.path.exists(array_subdir):
                os.makedirs(array_subdir)
                
            filenames = []
            for sample in samples:
                filename = self.get_ntuple_sample_path(sample_name=sample, dirname=ntuple_subdir)
                filenames.append(filename)

            args = (filenames, samples, repeat(self.treename), repeat(array_subdir),
                    repeat(columns), repeat(apply_columns), repeat(drop_columns),
                    repeat(fmt), repeat(downcast), repeat(complevel), repeat(chunksize),
                    repeat(library), repeat(kwargs))

            execute_multi_tasks(self._convert, *args, parallel=parallel)


def redirect_print_stream(streamer):
    import ROOT
    old_streamer = ROOT.RooPrintable.defaultPrintStream(streamer)
    return old_streamer

def get_default_stream():
    import ROOT
    return ROOT.RooPrintable.defaultPrintStream()


from .rooproc_define import RooProcDefine

class RooProcRedefine(RooProcDefine):
    
    def _execute(self, rdf, **params):
        name = params['name']
        expression = params['expression']
        if not hasattr(rdf, "Redefine"):
            raise RuntimeError("RDF.Redefine action requires ROOT version >= 6.26/00")
        rdf_next = rdf.Redefine(name, expression)
        return rdf_next


from typing import Dict, Optional, Union, List
import numpy as np
import pandas as pd

from quickstats.plots import AbstractPlot
from quickstats.plots.template import create_transform
from quickstats.utils.common_utils import combine_dict
from matplotlib.lines import Line2D


class Likelihood2DPlot(AbstractPlot):

    CONFIG = {
        'sigma_levels': ('1sigma', '2sigma', '3sigma'),
        'sigma_pos': 0.93,
        'sigma_names': ('1 $\sigma$', '2 $\sigma$', '3 $\sigma$'),
        'sigma_colors': ("hh:darkblue", "#F2385A", "#FDC536"),
        'highlight_styles': {
            'linewidth': 0,
            'marker': '*',
            'markersize': 20,
            'color': '#E9F1DF',
            'markeredgecolor': 'black'
        },
        'bestfit_styles':{
            'marker': 'P',
            'linewidth': 0,
            'markersize': 15
        },
        'bestfit_label': "Best fit ({x}, {y})",
        'cmap': 'GnBu',
        'interpolation': 'cubic',
        'num_grid_points': 100,
    }
    # https://pdg.lbl.gov/2018/reviews/rpp2018-rev-statistics.pdf#page=31
    likelihood_label_threshold = {
        '0.68': 2.30,
        '1sigma': 2.30,
        '0.90': 4.61,
        '0.95': 5.99,
        '2sigma': 6.18,
        '0.99': 9.21,
        '3sigma': 11.83,
    }

    def __init__(self, data_map: Union[pd.DataFrame, Dict[str, pd.DataFrame]],
                 label_map: Optional[Dict] = None,
                 styles_map: Optional[Dict] = None,
                 config_map: Optional[Dict] = None,
                 color_cycle=None,
                 styles: Optional[Union[Dict, str]] = None,
                 analysis_label_options: Optional[Dict] = None,
                 config: Optional[Dict] = None):

        self.data_map = data_map
        self.label_map = label_map
        self.styles_map = styles_map
        self.config_map = config_map
        self.highlight_data = []
        self.legend_order = []
        self.contours = None

        super().__init__(color_cycle=color_cycle,
                         styles=styles,
                         analysis_label_options=analysis_label_options,
                         config=config)

    def get_default_legend_order(self):
        if not isinstance(self.data_map, dict):
            return self.legend_order
        else:
            return list(self.data_map)

    def draw_single_data(self, ax, data: pd.DataFrame,
                         xattrib: str, yattrib: str, zattrib: str = 'qmu',
                         styles: Optional[Dict] = None,
                         config: Optional[Dict] = None,
                         clabel_size=None, show_colormesh=False):
        if config is None:
            config = self.config
        colors = list(config['sigma_colors'])
        if 'sigma_linestyles' not in config:
            config['sigma_linestyles'] = ['solid'] * \
                len(list(config['sigma_colors']))
        linestyles = list(config['sigma_linestyles'])
        levels = [self.likelihood_label_threshold[key]
                        for key in config['sigma_levels']]
        if config['interpolation'] is not None:
            from scipy import interpolate
            x = data[xattrib]
            y = data[yattrib]
            z = data[zattrib]

            def get_grid(X_range, Y_range, num_grid_points):
                X_points = np.linspace(*X_range, num_grid_points)
                Y_points = np.linspace(*Y_range, num_grid_points)
                X_grid, Y_grid = np.meshgrid(X_points, Y_points)
                return X_grid, Y_grid
            X, Y = get_grid(X_range=(x.min(), x.max()), Y_range=(
                y.min(), y.max()), num_grid_points=config['num_grid_points'])
            Z = interpolate.griddata(
                np.stack((x, y), axis=1), z, (X, Y), method=config['interpolation'])
        else:  # buggy for [dict]
            X_unique = np.sort(self.data_map[xattrib].unique())
            Y_unique = np.sort(self.data_map[yattrib].unique())
            X, Y = np.meshgrid(X_unique, Y_unique)
            Z = self.data_map.pivot_table(
                index=xattrib, columns=yattrib, values=zattrib).T.values - self.data_map[zattrib].min()

        if show_colormesh:
            cmap = config['cmap']
            im = ax.pcolormesh(X, Y, Z, cmap=cmap, shading='auto')
            import matplotlib.pyplot as plt
            plt.colorbar(im, ax=ax, **config['colorbar']).set_label(**config['colorbar_label'])
        if levels:
            cp = ax.contour(X, Y, Z, levels=levels, colors=colors,
                        linestyles=linestyles, linewidths=3)
            if clabel_size is not None:
                ax.clabel(cp, inline=True, fontsize=clabel_size)
        custom_handles = [Line2D([0], [0], color=color, linestyle=linestyle, lw=3, label=label) \
                          for color, key, label, linestyle in \
                          zip(config['sigma_colors'], config['sigma_levels'], config['sigma_names'], config['sigma_linestyles'])]
        self.update_legend_handles(
            dict(zip(config['sigma_names'], custom_handles)))
        self.legend_order.extend(config['sigma_names'])
        return custom_handles

    def draw(self, xattrib: str, yattrib: str, zattrib: str = 'qmu', xlabel: Optional[str] = "",
             ylabel: Optional[str] = "", zlabel: Optional[str] = "$-2\Delta ln(L)$",
             ymax: float = 5, ymin: float = -5, xmin: Optional[float] = -10, xmax: Optional[float] = 10,
             clabel_size=None, draw_sm_line: bool = False, draw_bestfit:bool=True,
             show_colormesh=False, show_legend=True):
        ax = self.draw_frame()
        self.contours = {'keys': [], 'contours': [], 'levels': []}
        if isinstance(self.data_map, pd.DataFrame):
            self.draw_single_data(ax, self.data_map, xattrib=xattrib, yattrib=yattrib,
                                  zattrib=zattrib, styles=self.styles_map, config=self.config,
                                  clabel_size=clabel_size, show_colormesh=show_colormesh)
        elif isinstance(self.data_map, dict):
            if self.styles_map is None:
                styles_map = {k: None for k in self.data_map}
            else:
                styles_map = self.styles_map
            if self.label_map is None:
                label_map = {k: k for k in self.data_map}
            else:
                label_map = self.label_map
            if self.config_map is None:
                config_map = {k: k for k in self.data_map}
            else:
                config_map = self.config_map
                
            for key in self.data_map:
                self.contours['keys'].append(key)
                data = self.data_map[key]
                styles = styles_map.get(key, None)
                label = label_map.get(key, "")
                config = config_map.get(key, None)
                config = combine_dict(self.CONFIG, config)
                handle = self.draw_single_data(ax, data,
                                               xattrib=xattrib,
                                               yattrib=yattrib,
                                               zattrib=zattrib,
                                               styles=styles,
                                               config=config,
                                               show_colormesh=show_colormesh)
        else:
            raise ValueError("invalid data format")
        
        highlight_index = 0
        
        if draw_bestfit and isinstance(self.data_map, pd.DataFrame):
            bestfit_idx = self.data_map[[zattrib]].idxmin()
            bestfit_x   = self.data_map[xattrib].iloc[bestfit_idx].values[0]
            bestfit_y   = self.data_map[yattrib].iloc[bestfit_idx].values[0]
            highlight_data = {
                'x': bestfit_x,
                'y': bestfit_y,
                'label': self.config["bestfit_label"].format(x=bestfit_x, y=bestfit_y),
                'styles': self.config["bestfit_styles"]
            }
            self.draw_highlight(ax, highlight_data, highlight_index)
            highlight_index += 1
        
        if self.highlight_data is not None:
            for i, h in enumerate(self.highlight_data):
                self.draw_highlight(ax, h, highlight_index + i)

        if draw_sm_line:
            sm_line_styles = self.config['sm_line_styles']
            sm_values = self.config['sm_values']
            transform = create_transform(
                transform_y="axis", transform_x="data")
            ax.vlines(sm_values[0], ymin=0, ymax=1, zorder=0, transform=transform,
                      **sm_line_styles)
            transform = create_transform(
                transform_x="axis", transform_y="data")
            ax.hlines(sm_values[1], xmin=0, xmax=1, zorder=0, transform=transform,
                      **sm_line_styles)

        if show_legend:
            handles, labels = self.get_legend_handles_labels()
            ax.legend(handles, labels, **self.styles['legend'])

        self.draw_axis_components(ax, xlabel=xlabel, ylabel=ylabel)
        self.set_axis_range(ax, xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax)

        return ax

    def draw_highlight(self, ax, data, index=0):
        styles = data['styles']
        if styles is None:
            styles = self.config['highlight_styles']
        handle = ax.plot(data['x'], data['y'], label=data['label'], **styles)
        self.update_legend_handles({f'highlight_{index}': handle[0]})
        self.legend_order.append(f'highlight_{index}')

    def add_highlight(self, x: float, y: float, label: str = "SM prediction",
                      styles: Optional[Dict] = None):
        highlight_data = {
            'x': x,
            'y': y,
            'label': label,
            'styles': styles
        }
        self.highlight_data.append(highlight_data)



from typing import Optional, Union, List, Dict

import numpy as np

from quickstats.maths.numerics import get_proper_ranges

def get_analysis_region_splitting_1D(value_range:List[float], signal_range:Optional[List[float]]=None,
                                     nominal_value:Optional[float]=None):
    if value_range is not None:
        value_range = get_proper_ranges(value_range, nominal_value)
        if value_range.shape != (1, 2):
            raise ValueError("value range must be list of size 2")
    if signal_range is not None:
        if value_range is None:
            raise RuntimeError("can not define signal range without defining value range")
        signal_range = get_proper_ranges(signal_range)
        if signal_range.shape != (1, 2):
            raise ValueError("signal range must be list of size 2")
        combined_range = np.array([[value_range[0][0], signal_range[0][0]],
                                   [signal_range[0][0], signal_range[0][1]],
                                   [signal_range[0][1], value_range[0][1]]])
        combined_range = get_proper_ranges(combined_range, no_overlap=True)
        return combined_range
    return value_range


from typing import List, Optional
import re

from quickstats import semistaticmethod, TVirtualNode, TVirtualTree
from quickstats.utils.string_utils import split_lines, split_str

from .actions import RooProcBaseAction, RooProcNestedAction, get_action

class ActionNode(TVirtualNode):
    
    def __init__(self, name:Optional[str]=None, level:Optional[int]=0,
                 parent:Optional["DomainNode"]=None,
                 **data):
        super().__init__(name=name, level=level,
                         parent=parent, **data)
        self.action = None
        
    def construct_action(self):
        action_cls = get_action(self.name)
        if action_cls is None:
            raise RuntimeError(f'unknown action "{self.name}"')        
        main_text  = self.get_data("main_text")
        block_text = self.get_data("block_text")
        action = action_cls.parse(main_text=main_text, block_text=block_text)
        self.action = action        
        
class ActionTree(TVirtualTree):
    
    NodeClass = ActionNode
    
    def construct_actions(self):
        self.reset()
        node = self.get_next()
        while node is not None:
            node.construct_action()
            node = self.get_next()
        self.reset()
    
class RooConfigLine:
    
    def __init__(self, text:str, line_number:int):
        self.text = text
        self.line_number = line_number
        self.start_tag = self.get_start_tag()
        if self.start_tag:
            self.end_tag = None
        else:
            self.end_tag = self.get_end_tag()
    
    def get_start_tag(self):
        result = re.search(r"^\s*<(?!/)([^>]+)>\s*$", self.text)
        if result:
            return result.group(1)
        return result
    
    def get_end_tag(self):
        result = re.search(r"^\s*</([^>]+)>\s*$", self.text)
        if result:
            if not re.search(r"^\s*</(\w+)>\s*$", self.text):
                raise ValueError(f'Line {self.line_number}: invalid end tag syntax "{self.text}"')
            return result.group(1)
        return result
    
class RooConfigParser(object):
    
    def __init__(self):
        pass

    @semistaticmethod
    def _get_action_tree(self, clines_iter, action_tree:Optional[ActionTree]=None):
        if action_tree is None:
            action_tree = ActionTree()
        cline = next(clines_iter, None)
        if cline is None:
            current_node = action_tree.current_node
            if current_node.name is not None:
                if current_node.data["end_line_number"] == -1:
                    raise RuntimeError(f'unterminated start tag "{current_node.data["raw_text"]}" '
                                       f'at Line {current_node.data["start_line_number"]}')
            return action_tree
        if cline.start_tag:
            current_node = action_tree.current_node
            if current_node.name is not None:
                action = get_action(current_node.name)
                if not issubclass(action, RooProcNestedAction):
                    raise RuntimeError(f'Line {cline.line_number}: can not create action block within an unnestable '
                                       f'action block "{current_node.name}" '
                                       f'(Line {current_node.data["start_line_number"]})')
            tokens = split_str(cline.start_tag)
            action_name = tokens[0]
            if len(tokens) > 1:
                block_text = " ".join(tokens[1:])
            else:
                block_text = None
            child_node = action_tree.add_child(action_name,
                                               raw_text=cline.text,
                                               start_line_number=cline.line_number,
                                               end_line_number=-1,
                                               block_text=block_text,
                                               main_text="")
            action_tree.current_node = child_node
        elif cline.end_tag:
            current_node = action_tree.current_node
            if current_node.name is None:
                raise RuntimeError(f'Line {cline.line_number}: found close tag '
                                   f'"{cline.text}" without a start tag')
            if current_node.name != cline.end_tag:
                raise RuntimeError(f'Line {cline.line_number}: close tag '
                                   f'"{cline.text}" does not match the start tag '
                                   f'"{current_node.data["raw_text"]}" (Line {current_node.data["start_line_number"]})')
            current_node.data["end_line_number"] = cline.line_number
            action_tree.current_node = current_node.parent
        else:
            current_node = action_tree.current_node
            action = None
            if current_node.name is not None:
                action = get_action(current_node.name)
            if action and not issubclass(action, RooProcNestedAction):
                current_node.data["main_text"] += cline.text
            else:
                tokens = split_str(cline.text)
                action_name = tokens[0]
                if len(tokens) > 1:
                    main_text = " ".join(tokens[1:])
                else:
                    main_text = None
                child_node = action_tree.add_child(action_name,
                                                   raw_text=cline.text,
                                                   start_line_number=cline.line_number,
                                                   end_line_number=cline.line_number,
                                                   main_text=main_text,
                                                   block_text=None)
        return self._get_action_tree(clines_iter, action_tree)
        
    @semistaticmethod
    def parse_file(self, path:str):
        with open(path, "r") as f:
            text = f.read()
        return self.parse_text(text)        

    @semistaticmethod
    def parse_text(self, text:str):
        numbered_lines = split_lines(text, comment_string="#", remove_blank=True, with_line_number=True)
        clines = [RooConfigLine(line, line_number) for line, line_number in numbered_lines]
        clines_iter = iter(clines)
        action_tree = self._get_action_tree(clines_iter)
        return action_tree
    


from typing import Dict, List, Optional, Union, Tuple
import os
import copy
import json

from quickstats import ConfigurableObject, ConfigComponent

from .analysis_path_manager import AnalysisPathManager
from .config_format_templates import DEFAULT_ANALYSIS_CONFIG_FORMAT

class AnalysisBase(ConfigurableObject):
    
    CONFIG_FORMAT = DEFAULT_ANALYSIS_CONFIG_FORMAT
    
    REQUIRED_CONFIG_COMPONENTS = []
    
    @property
    def names(self) -> Dict[str, str]:
        try:
            names = self.config["names"]
        except:
            names = {}
        return names
    
    def __init__(self, analysis_config:Optional[Union[Dict, str]]=None,
                 plot_config:Optional[Union[Dict, str]]=None,
                 outdir:Optional[str]=None,
                 ntuple_dir:Optional[str]=None,
                 array_dir:Optional[str]=None,
                 model_dir:Optional[str]=None,
                 path_manager:Optional[AnalysisPathManager]=None,
                 disable_config_message:bool=False,
                 verbosity:Optional[Union[int, str]]="INFO"):
        
        super().__init__(disable_config_message=disable_config_message,
                         verbosity=verbosity)
        
        if path_manager is None:
            self.path_manager = AnalysisPathManager()
        else:
            self.path_manager = path_manager
        
        self.load_analysis_config(analysis_config)
        self.load_plot_config(plot_config)
        
        paths = {
            "outdir": outdir,
            "ntuple_dir": ntuple_dir,
            "array_dir": array_dir,
            "model_dir": model_dir
        }
        for dirname, config_key in [("outdir", "outputs"),
                                    ("ntuple_dir", "ntuples"),
                                    ("array_dir", "arrays"),
                                    ("model_dir", "models")]:
            # use custom path given by the user
            if paths[dirname] is not None:
                continue
            # already defined in path manager
            if (dirname == "outdir") and (self.path_manager.base_path is not None):
                continue
            elif (dirname.replace("_dir", "") in self.path_manager.directories):
                continue
            # use path defined in config file
            try:
                paths[dirname] = self.config["paths"][config_key]
            # use path defined in the default config format
            except:
                paths[dirname] = DEFAULT_ANALYSIS_CONFIG_FORMAT["paths"][config_key].default
            
        # setup file paths used in the analysis pipeline
        self.update_paths(**paths)
                                        
    def update_paths(self, outdir:Optional[str]=None,
                     directories:Optional[Dict[str, str]]=None,
                     files:Optional[Dict[str, Union[str, Tuple[Optional[str], str]]]]=None,
                     ntuple_dir:Optional[str]=None,
                     array_dir:Optional[str]=None,
                     model_dir:Optional[str]=None):
        
        if outdir is not None:
            self.path_manager.set_base_path(outdir)
        if directories is not None:
            self.path_manager.update_directories(directories)
        if files is not None:
            self.path_manager.update_files(files)
        if ntuple_dir is not None:
            self.path_manager.set_directory("ntuple", ntuple_dir, absolute=True)
        if array_dir is not None:
            self.path_manager.set_directory("array", array_dir, absolute=True)
        if model_dir is not None:
            self.path_manager.set_directory("model", model_dir, absolute=True)
            
    def set_study_name(self, study_name:str):
        self.path_manager.set_study_name(study_name)
        if "model" in self.path_manager.directories:
            model_dir = self.path_manager.directories["model"]
            basename = os.path.basename(os.path.dirname(model_dir))
            if basename != study_name:
                model_dir = os.path.join(model_dir, study_name)
                self.path_manager.set_directory("model", model_dir)
        
    def get_study_name(self):
        return self.path_manager.study_name
    
    def get_directory(self, directory_name:str):
        return self.path_manager.get_directory(directory_name)
    
    def get_file(self, file_name:str, validate:bool=False, **parameters):
        if validate:
            self._check_file(file_name, **parameters)
        return self.path_manager.get_resolved_file(file_name, **parameters)
    
    def _has_directory(self, directory_name:str):
        return self.path_manager.directory_exists(directory_name)
    
    def _has_file(self, file_name:str, **parameters):
        return self.path_manager.file_exists(file_name, **parameters)
    
    def _check_directory(self, directory_name:str):
        self.path_manager.check_directory(directory_name)
    
    def _check_file(self, file_name:str, **parameters):
        self.path_manager.check_file(file_name, **parameters)
        
    def load_analysis_config(self, config_source:Optional[Union[Dict, str]]=None):
        if isinstance(config_source, str):
            if not os.path.exists(config_source):
                raise FileNotFoundError(f'config file "{config_source}" does not exist')
            config_path = os.path.abspath(config_source)
            self.path_manager.set_file("analysis_config", config_path)
        if config_source is None:
            self.load_config({})
        else:
            self.load_config(config_source)
        try:
            self.all_channels = list(self.config['channels'])
        except:
            self.all_channels = []
        try:
            self.all_kinematic_regions = list(self.config['kinematic_regions'])
        except:
            self.all_kinematic_regions = []
        try:
            self.all_samples = list(self.config['samples']['all'])
        except:
            self.all_samples = []
        try:
            self.extra_samples = list(self.config['samples']['extra'])
        except:
            self.extra_samples = []            
        try:
            self.all_variables = list(self.config['variables']['all'])
        except:
            self.all_variables = []
        try:
            self.treename = self.config['names']['tree_name']
        except:
            self.treename = DEFAULT_CONFIG_FORMAT["names"]["tree_name"].default
        
    def load_plot_config(self, config_source:Optional[Union[Dict, str]]=None):
        if isinstance(config_source, str):
            if not os.path.exists(config_source):
                raise FileNotFoundError(f'config file "{config_source}" does not exist')
            config_path = os.path.abspath(config_source)
            self.path_manager.set_file("plot_config", config_path)
        # use the default plot config from the framework
        if config_source is None:
            self.plot_config = {}
            return None
        elif isinstance(config_source, str):
            with open(config_source, "r") as file:
                self.plot_config = json.load(file)
        elif isinstance(config_source, dict):
            self.plot_config = copy.deepcopy(config_source)
        else:
            raise RuntimeError("invalid plot config format")  
        
    def resolve_channels(self, channels:List[str]):
        for channel in channels:
            if channel not in self.all_channels:
                raise ValueError(f"unknown channel: {channel}")
        return channels
    
    def resolve_samples(self, samples:Optional[List[str]]=None):
        if samples is None:
            return self.all_samples
        resolved_samples = []
        for sample_key in samples:
            if sample_key in self.config['samples']:
                for sample in self.config['samples'][sample_key]:
                    if sample not in resolved_samples:
                        resolved_samples.append(sample)
            elif (sample_key in self.all_samples) or (sample_key in self.extra_samples):
                resolved_samples.append(sample_key)
            else:
                raise RuntimeError(f"unknown sample \"{sample_key}\"")
        return resolved_samples

    def resolve_variables(self, variables:Optional[List[str]]=None):
        if variables is None:
            return self.all_variables
        resolved_variables = []
        for variable_key in variables:
            if variable_key in self.config['variables']:
                for variable in self.config['variables'][variable_key]:
                    if variable not in resolved_variables:
                        resolved_variables.append(variable)
            elif variable_key in self.all_variables:
                resolved_variables.append(variable_key)
            else:
                raise RuntimeError(f"unknown variable \"{variable_key}\"")
        return resolved_variables
    
    def resolve_class_labels(self, class_labels:Dict):
        resolved_class_labels = {}
        for label in class_labels:
            resolved_samples = self.resolve_samples(class_labels[label])
            for sample in resolved_samples:
                if sample not in resolved_class_labels:
                    resolved_class_labels[sample] = label
                else:
                    raise RuntimeError(f"multiple class labels found for the sample \"{sample}\"")
        return resolved_class_labels
    
    def get_analysis_data_format(self):
        try:
            fmt = self.config["data_storage"]["analysis_data_arrays"]["storage_format"]
        except:
            fmt = DEFAULT_CONFIG_FORMAT["data_storage"]["analysis_data_arrays"]["storage_format"]
        return fmt
    
    def get_event_index_variable(self):
        if "event_number" not in self.names:
            raise RuntimeError('no event index variable defined')
        index_variable = self.names["event_number"]
        return index_variable
    
    def get_blind_sample_name(self, sample:str):
        return f"{sample}_blind"
    
    def is_data_sample(self, sample:str):
        """Check whether a given sample is the observed data. Analysis specific, should be overridden.
        """
        return "data" in sample


from typing import Optional, List, Dict
import os
import json

from .rooproc_helper_action import RooProcHelperAction

class RooProcExport(RooProcHelperAction):
    def __init__(self, filename:str):
        super().__init__(filename=filename)
        
    @classmethod
    def parse(cls, main_text:str, block_text:Optional[str]=None):
        kwargs = cls.parse_as_kwargs(main_text)
        return cls(**kwargs)
    
    def _execute(self, processor:"quickstats.RooProcessor", **params):
        filename = params['filename']
        data = {k:v.GetValue() for k,v in processor.external_variables.items()}
        dirname = os.path.dirname(filename)
        if dirname and (not os.path.exists(dirname)):
            os.makedirs(dirname)
        with open(filename, 'w') as outfile:
            processor.stdout.info(f'INFO: Writing auxiliary data to "{filename}".')
            json.dump(data, outfile, indent=2)
        return processor


from typing import List, Optional

from quickstats.utils.root_utils import declare_expression

ROOT_MACROS = \
{
    "TH1Utils":
    """
    namespace TH1Utils {
        template<typename T>
        std::vector<T> GetBinErrorArray(TH1* h, const size_t &underflow_bin=0, const size_t & overflow_bin=0)
        {
            const size_t n_bin = h->GetNbinsX();
            std::vector<T> result;
            result.reserve(n_bin);
            const size_t bin_min = 1 - underflow_bin;
            const size_t bin_max = n_bin + overflow_bin;
            for (size_t bin_index = bin_min; bin_index <= bin_max; bin_index++)
                result.push_back(h->GetBinError(bin_index));
            return result;
        }

        template<typename T>
        std::vector<T> GetBinErrorUpArray(TH1* h, const size_t &underflow_bin=0, const size_t & overflow_bin=0)
        {
            const size_t n_bin = h->GetNbinsX();
            std::vector<T> result;
            result.reserve(n_bin);
            const size_t bin_min = 1 - underflow_bin;
            const size_t bin_max = n_bin + overflow_bin;
            for (size_t bin_index = bin_min; bin_index <= bin_max; bin_index++)
                result.push_back(h->GetBinErrorUp(bin_index));
            return result;
        }

        template<typename T>
        std::vector<T> GetBinErrorLowArray(TH1* h, const size_t &underflow_bin=0, const size_t & overflow_bin=0)
        {
            const size_t n_bin = h->GetNbinsX();
            std::vector<T> result;
            result.reserve(n_bin);
            const size_t bin_min = 1 - underflow_bin;
            const size_t bin_max = n_bin + overflow_bin;
            for (size_t bin_index = bin_min; bin_index <= bin_max; bin_index++)
                result.push_back(h->GetBinErrorLow(bin_index));
            return result;
        }

        template<typename T>
        std::vector<T> GetBinCenterArray(TH1* h, const size_t &underflow_bin=0, const size_t & overflow_bin=0)
        {
            const size_t n_bin = h->GetNbinsX();
            std::vector<T> result;
            result.reserve(n_bin);
            const size_t bin_min = 1 - underflow_bin;
            const size_t bin_max = n_bin + overflow_bin;
            for (size_t bin_index = bin_min; bin_index <= bin_max; bin_index++)
                result.push_back(h->GetBinCenter(bin_index));
            return result;
        }

        template<typename T>
        std::vector<T> GetBinContentArray(TH1* h, const size_t &underflow_bin=0, const size_t & overflow_bin=0)
        {
            const size_t n_bin = h->GetNbinsX();
            std::vector<T> result;
            result.reserve(n_bin);
            const size_t bin_min = 1 - underflow_bin;
            const size_t bin_max = n_bin + overflow_bin;
            for (size_t bin_index = bin_min; bin_index <= bin_max; bin_index++)
                result.push_back(h->GetBinContent(bin_index));
            return result;
        }
      
        template<typename T>
        std::vector<T> GetBinWidthArray(TH1* h, const size_t &underflow_bin=0, const size_t & overflow_bin=0)
        {
            const size_t n_bin = h->GetNbinsX();
            std::vector<T> result;
            result.reserve(n_bin);
            const size_t bin_min = 1 - underflow_bin;
            const size_t bin_max = n_bin + overflow_bin;
            for (size_t bin_index = bin_min; bin_index <= bin_max; bin_index++)
                result.push_back(h->GetBinWidth(bin_index));
            return result;
        }

        template<typename T>
        std::vector<T> GetBinLowEdgeArray(TH1* h, const size_t &underflow_bin=0, const size_t & overflow_bin=0)
        {
            const size_t n_bin = h->GetNbinsX();
            std::vector<T> result;
            result.reserve(n_bin);
            const size_t bin_min = 1 - underflow_bin;
            // the + 1 is because number of bin low edges = number of bins + 1
            const size_t bin_max = n_bin + overflow_bin + 1;
            for (size_t bin_index = bin_min; bin_index <= bin_max; bin_index++)
                result.push_back(h->GetBinLowEdge(bin_index));
            return result;
        }
        
        template<typename T>
        int FindBinIndexByBinEdge(TH1* h, const T &bin_edge, const double &epsilon=1e-6)
        {
            const T first_edge = h->GetBinLowEdge(1);
            if ((bin_edge < first_edge) && abs(bin_edge - first_edge) > epsilon)
                return 0;
            auto n_bins = h->GetNbinsX();
            const T last_edge = h->GetBinLowEdge(n_bins);
            if ((bin_edge > last_edge) && abs(bin_edge - last_edge) > epsilon)
                return n_bins + 1;
            for (size_t i = 1; i < n_bins + 1; i++){
                const T edge_i = h->GetBinLowEdge(i);
                if (abs(bin_edge - edge_i) < epsilon)
                    return i;
            }
            return -1;
        }
    }; 
    """,
    "TF1Utils":
    """
    namespace TF1Utils {
        std::vector<double> GetRandomArray(TF1* f, const int &size, const double &xmin, const double &xmax){
            std::vector<double> result(size);
            for (size_t i = 0; i < size; i++)
                result[i] = f->GetRandom(xmin, xmax);
            return result;
        }
        std::vector<double> GetRandomArray(TF1* f, const int &size){
            std::vector<double> result(size);
            for (size_t i = 0; i < size; i++)
                result[i] = f->GetRandom();
            return result;
        }
    }; 
    """,    
    "TAxisUtils": 
    """
    namespace TAxisUtils{
        template<typename T>
        std::vector<T> GetBinLowEdgeArray(TAxis* ax, const size_t &underflow_bin=0, const size_t & overflow_bin=0)
        {
            const size_t n_bin = ax->GetNbins();
            std::vector<T> result;
            result.reserve(n_bin);
            const size_t bin_min = 1 - underflow_bin;
            const size_t bin_max = n_bin + overflow_bin;
            for (size_t bin_index = bin_min; bin_index <= bin_max; bin_index++)
                result.push_back(ax->GetBinLowEdge(bin_index));
            return result;
        }
        template<typename T>
        std::vector<T> GetBinCenterArray(TAxis* ax, const size_t &underflow_bin=0, const size_t & overflow_bin=0)
        {
            const size_t n_bin = ax->GetNbins();
            std::vector<T> result;
            result.reserve(n_bin);
            const size_t bin_min = 1 - underflow_bin;
            const size_t bin_max = n_bin + overflow_bin;
            for (size_t bin_index = bin_min; bin_index <= bin_max; bin_index++)
                result.push_back(ax->GetBinCenter(bin_index));
            return result;
        }
        template<typename T>
        std::vector<T> GetBinWidthArray(TAxis* ax, const size_t &underflow_bin=0, const size_t & overflow_bin=0)
        {
            const size_t n_bin = ax->GetNbins();
            std::vector<T> result;
            result.reserve(n_bin);
            const size_t bin_min = 1 - underflow_bin;
            const size_t bin_max = n_bin + overflow_bin;
            for (size_t bin_index = bin_min; bin_index <= bin_max; bin_index++)
                result.push_back(ax->GetBinWidth(bin_index));
            return result;
        }
    };
    """,    
    "THistUtils": 
    """
    namespace THistUtils{
        template<typename T>
        std::vector<double> GetPoissonError(const std::vector<T> data, const double& nSigma=1){
            Double_t ym, yp;
            auto inst = RooHistError::instance();
            const int data_size = data.size();
            std::vector<double> result(2*data_size);
            for (size_t i = 0; i < data_size; i++){
                inst.getPoissonInterval(Int_t(data[i] + 0.5), ym, yp, nSigma);
                result[i] = data[i] - ym;
                result[data_size + i] = yp - data[i];
            }
            return result;
        }
    };
    """,
    "RFUtils":
    """
    namespace RFUtils{
        struct DatasetStruct {
          std::vector<double> observable_values;
          std::vector<double> weights;
          std::vector<std::string> category_labels;
          std::vector<int> category_index;
          DatasetStruct(const size_t &n_entries, const size_t &n_obs){
              observable_values = std::vector<double>(n_obs * n_entries);
              weights = std::vector<double>(n_entries);
              category_labels = std::vector<std::string>(n_entries);
              category_index = std::vector<int>(n_entries);
          }
        } ;
        
        DatasetStruct ExtractCategoryData(const RooAbsData* dataset, const RooArgSet* observables, const RooCategory* cat){
            const size_t n_entries = dataset->numEntries();
            const size_t n_obs = observables->size();
            DatasetStruct result(n_entries, n_obs);
            TIterator* iter(observables->createIterator());
            RooRealVar* obs;
            for (size_t i = 0; i < n_entries; i++){
                dataset->get(i);
                for (size_t j = 0; j < n_obs; j++)
                    result.observable_values[i + j * n_entries] = ((RooRealVar*)(*observables)[j])->getVal();
                result.weights[i] = dataset->weight();
                result.category_labels[i] = cat->getLabel();
                result.category_index[i] = cat->getIndex();
            }
            return result;
        }
        
        DatasetStruct ExtractData(const RooAbsData* dataset, const RooArgSet* observables){
            const size_t n_entries = dataset->numEntries();
            const size_t n_obs = observables->size();
            DatasetStruct result(n_entries, n_obs);
            TIterator* iter(observables->createIterator());
            RooRealVar* obs;
            for (size_t i = 0; i < n_entries; i++){
                dataset->get(i);
                for (size_t j = 0; j < n_obs; j++)
                    result.observable_values[i + j * n_entries] = ((RooRealVar*)(*observables)[j])->getVal();
                result.weights[i] = dataset->weight();
            }
            return result;
        }
        
         void CopyData(const RooAbsData* source, RooAbsData* target, const RooRealVar* source_var,
                       RooRealVar* target_var, RooRealVar* weight){
             for (size_t i = 0; i < source->numEntries(); i++){
                 source->get(i);
                 double weight_val = source->weight();
                 weight->setVal(weight_val);
                 target_var->setVal(source_var->getVal());
                 target->add(RooArgSet(*target_var, *weight), weight_val);
             }
        }
        
        struct RooArgSetInfo {
          std::vector<std::string> class_names;
          std::vector<std::string> names;
          RooArgSetInfo(const RooArgSet* argset){
            const size_t n_obj = argset->size();
            this->class_names = std::vector<std::string>(n_obj);
            this->names = std::vector<std::string>(n_obj);
            for (size_t i = 0; i < n_obj; i++){
                this->class_names[i] = ((TObject*)(*argset)[i])->ClassName();
                this->class_names[i] = ((TObject*)(*argset)[i])->GetName();
            }
          }
        } ;
        
        #if ROOT_VERSION_CODE >= ROOT_VERSION(6,26,0)
        #endif
    };
    """
}

def load_macro(macro_name:str):
    expression = ROOT_MACROS.get(macro_name, None)
    if expression is None:
        raise ValueError(f"`{macro_name}` is not a built-in quickstats macro."
                         " Available macros are: {}".format(",".join(list(ROOT_MACROS))))
    declare_expression(expression, macro_name)

def load_macros(macro_names:Optional[List[str]]=None):
    if macro_names is None:
        macro_names = list(ROOT_MACROS)
    for macro_name in macro_names:
        load_macro(macro_name)


from typing import Optional, Union, List, Dict

import numpy as np

from quickstats import DescriptiveEnum

class NegativeWeightMode(DescriptiveEnum):
    DONOTHING = (0, "do not apply any correction to negative weight")
    SETZERO   = (1, "set negative weight to 0")
    SETABS    = (2, "set negative weight to its absolute value")

def fix_negative_weights(df, mode:Union[int, NegativeWeightMode, str]=0,
                         weight_col:str="weight"):
    mode = NegativeWeightMode.parse(mode)
    if mode == NegativeWeightMode.DONOTHING:
        return None
    mask = df[weight_col] < 0
    if mode == NegativeWeightMode.SETZERO:
        df[weight_col][mask] = 0
    elif mode == NegativeWeightMode.SETABS:
        df[weight_col][mask] = abs(df[weight_col][mask])
        
def shuffle_arrays(*arrays, random_state:Optional[int]=None):
    if random_state < 0:
        return arrays
    from sklearn.utils import shuffle
    return shuffle(*arrays, random_state=random_state)

def get_class_label_encodings(class_labels:Union[List, np.ndarray]):
    from sklearn import preprocessing
    encodings = {}
    labelencoder = preprocessing.LabelEncoder()
    labelencoder.fit_transform(class_labels)
    encodings = dict(zip(labelencoder.classes_, np.arange(len(class_labels))))
    return encodings


from typing import Optional, Union, List, Dict, Callable

import ROOT

from quickstats import AbstractObject, DescriptiveEnum, PathManager
from quickstats.interface.root import RooRealVar

from .component_source import ComponentSource

class TreeDataSource(ComponentSource):
    """
    Handling Tree data from a ROOT file in the form of a RooDataSet.
    """
    
    _REQUIRE_CONFIG_ = {
        "ROOT"  : False,
        "RooFit": True
    }    

    def __init__(self, tree_name:str,
                 input_paths:Optional[Union[PathManager, str, List[str]]],
                 observable:Optional[Union[RooRealVar, ROOT.RooRealVar, dict, str]]=None,
                 weight_name:Optional[str]=None,
                 verbosity:Optional[Union[int, str]]="INFO"):
        super().__init__(input_paths=input_paths,
                         observable=observable,
                         verbosity=verbosity)
        self.weight_name  = weight_name
        self.tree_name    = tree_name
        
    def get_tree(self, **file_kwargs):
        if self.tree_name is None:
            raise RuntimeError("tree name not set")
        chain = ROOT.TChain(self.tree_name)
        input_paths = self.get_resolved_input_paths(**file_kwargs)
        for path in input_paths:
            chain.Add(path)
        return chain
        
    def create_dataset_from_tree(self, name:str, tree:Union[ROOT.TTree, ROOT.TChain]):
        """
        """
        variables = ROOT.RooArgSet()
        if self.observable is None:
            raise RuntimeError("observable not set")
        observable = self.observable.new()
        variables.add(observable)
        if self.weight_name is not None:
            weight_var = ROOT.RooRealVar(self.weight_name, self.weight_name, 1)
            variables.add(weight_var)
            dataset = ROOT.RooDataSet(name, name, variables,
                                      ROOT.RooFit.Import(tree),
                                      ROOT.RooFit.WeightVar(weight_var))
        else:
            dataset = ROOT.RooDataSet(name, name, variables,
                                      ROOT.RooFit.Import(tree))
        return dataset
    
    def create_dataset(self, name:str, **file_kwargs):
        """
        """
        tree = self.get_tree(**file_kwargs)
        dataset = self.create_dataset_from_tree(name, tree)
        return dataset
    
    def create_histogram(self, name:Optional[str]=None, n_bins:Optional[int]=None,
                         bin_range:Optional[List[float]]=None, **file_kwargs):
        tree = self.get_tree(**file_kwargs)
        if name is None:
            import uuid
            h_name = uuid.uuid4().hex
        else:
            h_name = name
        obs = self.observable.new()
        obs_name = obs.GetName()
        if n_bins is None:
            n_bins = obs.numBins()
        if bin_range is None:
            vmin, vmax = obs.getMin(), obs.getMax()
        else:
            vmin, vmax = bin_range[0], bin_range[1]
        if (vmin == -1e30) and (vmax == 1e30):
            self.stdout.warning("WARNING: Observable range is unbounded (-1e30, 1e30). "
                                "The resulting histogram could be buggy.")
        hist = ROOT.TH1D(h_name, h_name, n_bins, vmin, vmax)
        canvas = ROOT.TCanvas(uuid.uuid4().hex)
        if self.weight_name is not None:
            tree.Draw(f"{obs_name} >> {h_name}", self.weight_name)
        else:
            tree.Draw(f"{obs_name} >> {h_name}")
        del canvas
        hist.SetDirectory(0)
        return hist


from typing import Any
from enum import Enum

class GeneralEnum(Enum):
    
    __aliases__ = {
    }

    @classmethod
    def on_parse_exception(cls, expr:str):
        options = cls.get_members()
        raise RuntimeError(f"invalid option \"{expr}\" for the enum class \"{cls.__name__}\" "
                           f"(allowed options: {', '.join(options)})")
    
    @classmethod
    def parse(cls, expr:str):
        if isinstance(expr, cls):
            return expr
        elif isinstance(expr, int):
            return cls(expr)
        _expr = expr.strip().lower()
        members = cls.get_members_map()
        if _expr in members:
            return cls(members[_expr])
        else:
            # check aliases
            aliases = cls.get_aliases_map()
            if _expr in aliases:
                return cls(cls.parse(aliases[_expr]))
        cls.on_parse_exception(expr)
            
    @classmethod
    def get_members(cls):
        return [i.lower() for i in cls.__members__]
    
    @classmethod
    def get_members_map(cls):
        return {k.lower():v for k, v in cls.__members__.items()}
    
    @classmethod
    def get_aliases_map(cls):
        return {k.lower():v for k, v in cls.__aliases__.items()}
    
    @classmethod
    def has_member(cls, name:str):
        return name in cls.__members__
    
    @classmethod
    def get_member_by_attribute(cls, attribute:str, value:Any):
        members = cls.__members__
        return next((x for x in members.values() if getattr(x, attribute) == value), None)
    
class DescriptiveEnum(GeneralEnum):
    
    def __new__(cls, value:int, description:str=""):
        obj = object.__new__(cls)
        obj._value_ = value
        obj.description = description
        return obj

    @classmethod
    def on_parse_exception(cls, expr:str):
        enum_descriptions = "".join([f"    {key.lower()} - {val.description}\n" \
                                     for key, val in cls.__members__.items()])
        raise RuntimeError(f"invalid option \"{expr}\" for the enum class \"{cls.__name__}\"\n"
                           f"  Allowed options:\n{enum_descriptions}")


from typing import Optional, Tuple, List, Dict
import uuid
import array
import numpy as np

from quickstats.interface.root import TObject, TArrayData
from quickstats.interface.root import load_macro
from quickstats.interface.cppyy.vectorize import as_vector, as_np_array, as_c_array, np_type_str_maps

class TH1(TObject):
    
    def __init__(self, h:Optional["ROOT.TH1"]=None, dtype:str='double', 
                 underflow_bin:int=0,
                 overflow_bin:int=0):
        self.dtype = dtype
        self.underflow_bin = underflow_bin
        self.overflow_bin  = overflow_bin
        self.init(h)
        
    def get_fundamental_type(self):
        import ROOT
        return ROOT.TH1
    
    def sanity_check(self):
        if any(attribute is None for attribute in [self.bin_content, self.bin_error, self.bin_center,
                                                   self.bin_width, self.bin_low_edge]):
            raise RuntimeError("histogram not initialized")
        
    def init(self, h:Optional["ROOT.TH1"]=None):
        if h is not None:
            self.bin_content  = self.GetBinContentArray(h, self.dtype, self.underflow_bin, self.overflow_bin)
            self.bin_error    = self.GetBinErrorArray(h, self.dtype, self.underflow_bin, self.overflow_bin)
            self.bin_center   = self.GetBinCenterArray(h, self.dtype, self.underflow_bin, self.overflow_bin)
            self.bin_width    = self.GetBinWidthArray(h, self.dtype, self.underflow_bin, self.overflow_bin)
            self.bin_low_edge = self.GetBinLowEdgeArray(h, self.dtype, self.underflow_bin, self.overflow_bin)
        else:
            self.bin_content = None
            self.bin_error = None
            self.bin_center = None
            self.bin_width = None
            self.bin_low_edge = None
    
    @classmethod
    def from_numpy_data(cls, x:np.ndarray, bins:int=10, range:Optional[Tuple[float]]=None,
                        density:bool=False, weights:Optional[np.ndarray]=None, **kwargs):
        hist, bin_edges = np.histogram(x, bins=bins, range=range, density=density, weights=weights)
        return cls.from_numpy_histogram(hist, bin_edges, **kwargs)
    
    @classmethod
    def from_numpy_histogram(cls, hist:np.ndarray, bin_edges:np.ndarray, **kwargs):
        assert (hist.ndim == 1) and (bin_edges.ndim == 1)
        th1 = TH1(**kwargs)
        th1.bin_content = np.array(hist)
        th1.bin_error = np.zeros(th1.bin_content.shape)
        th1.bin_low_edge = np.array(bin_edges)
        from quickstats.maths.statistics import bin_edge_to_bin_center
        th1.bin_center = bin_edge_to_bin_center(th1.bin_low_edge)
        th1.bin_width = (th1.bin_low_edge[-1] - th1.bin_low_edge[0]) / (th1.bin_low_edge.shape[0] - 1)
        return th1
        
    @staticmethod
    def _to_ROOT(name:str, bin_content, bin_width=None, bin_center=None, bin_low_edge=None,
                 bin_error=None, title:str=None):
        import ROOT
        if ((bin_low_edge is None) and (bin_center is None)) or \
           ((bin_low_edge is not None) and (bin_center is not None)):
            raise ValueError("must provide either bin low edge or bin center values")
        n_bins = len(bin_content)
        if bin_center is not None:
            if bin_width is None:
                if n_bins < 2:
                    raise ValueError("unable to determine bin width with only one bin")
                bin_width = (bin_center[1] - bin_center[0])
            bin_low_edge = bin_center - bin_width/2
        if len(bin_low_edge) == n_bins:
            bin_width = (bin_low_edge[-1] - bin_low_edge[-2])
            bin_low_edge = np.insert(bin_low_edge, len(bin_low_edge), bin_low_edge[-1] + bin_width)
        if title is None:
            title = name
        dtype = bin_content.dtype
        if np_type_str_maps[dtype] == "float":
            h_func = ROOT.TH1F
        else:
            h_func = ROOT.TH1D
        h = h_func(name, title, n_bins, as_c_array(bin_low_edge))
        if bin_error is None:
            bin_error = np.zeros((n_bins,))
        for i in range(n_bins):
            h.SetBinContent(i + 1, bin_content[i])
            h.SetBinError(i + 1, bin_error[i])
        return h
    
    def to_ROOT(self, name:str, title:str=None):
        self.sanity_check()
        if title is None:
            title = name
        return self._to_ROOT(name, self.bin_content, bin_low_edge=self.bin_low_edge,
                             bin_error=self.bin_error, title=title)
    
    @staticmethod
    def find_bin_by_edge(h:"ROOT.TH1", low_edge:float, tol:float=1e-6):
        import ROOT
        return ROOT.TH1Utils.FindBinIndexByBinEdge(h, low_edge, tol)
    
    @staticmethod
    def find_binnings_relative_to_variable(h:"ROOT.TH1", variable:"ROOT.RooRealVar", tol:float=1e-6):
        vmin, vmax = variable.getMin(), variable.getMax()
        bin_low  = TH1.find_bin_by_edge(h, vmin, tol=tol)
        bin_high = TH1.find_bin_by_edge(h, vmax, tol=tol)
        n_bins   = bin_high - bin_low
        return bin_low, bin_high, n_bins
    
    
    @staticmethod
    def rebin_to_variable(h:"ROOT.TH1", variable:"ROOT.RooRealVar", tol:float=1e-6):
        bin_low, bin_high, n_bins = TH1.find_binnings_relative_to_variable(h, variable, tol=tol)
        var_n_bins = variable.getBins()
        if n_bins != var_n_bins:
            histname = h.GetName()
            varname = variable.GetName()
            if n_bins == 0:
                raise RuntimeError(f"Input histogram `{histname}` has 0 bins in the "
                                   f"range defined by the variable `{varname}`."
                                   f"Please check compatibility between your "
                                   f"variable and the histogram range")
            if (n_bins < var_n_bins):
                raise RuntimeError(f"input histogram `{histname}` has fewer bins "
                                   f"({n_bins}) than the variable `{varname}` "
                                   f"({var_n_bins})")
            if (n_bins > var_n_bins) and (n_bins % var_n_bins != 0):
                raise RuntimeError(f"input histogram `{histname}` has inconsistent "
                                   f"number of bins ({n_bins}) with the "
                                   f"variable `{varname}` ({var_n_bins})")
            else:
                ratio = n_bins // obs_n_bins
                h.Rebin(ratio)
                return ratio
        return 1
    
    @staticmethod
    def copy_histogram_in_effective_range(source:"ROOT.TH1", target:"ROOT.TH1",
                                          tol:float=1e-6, weight_scale:float=1):
        source_bin_edges = TH1.GetBinLowEdgeArray(source)
        target_bin_edges = TH1.GetBinLowEdgeArray(target)
        idx_first = np.where(np.abs(source_bin_edges - target_bin_edges[0]) < tol)[0]
        idx_last = np.where(np.abs(source_bin_edges - target_bin_edges[-1]) < tol)[0]
        if (((not idx_first) or (not idx_last)) or 
            (not np.allclose(source_bin_edges[idx_first[0]:idx_last[0]+1], target_bin_edges, atol=tol))):
            raise RuntimeError("binnings of the target histogram is not a subset of the "
                               "binnings of the source histogram")
        idx_first, idx_last = int(idx_first[0]), int(idx_last[0])
        n_bins = target.GetNbinsX()
        for i in range(1, n_bins + 1):
            bin_index = idx_first + i
            bin_content = source.GetBinContent(bin_index) * weight_scale
            bin_error = source.GetBinError(bin_index) * weight_scale
            target.SetBinContent(i, bin_content)
            target.SetBinError(i, bin_error)
    
    @staticmethod
    def apply_blind_range(h:"ROOT.TH1", blind_range:List[float]):
        assert len(blind_range) == 2
        n_bins = h.GetNbinsX()
        for i in range(1, n_bins + 1):
            bin_center = h.GetBinCenter(i)
            if (bin_center > blind_range[0]) and (bin_center < blind_range[1]):
                h.SetBinContent(i, 0)
    
    @staticmethod
    def remove_negative_bins(h:"ROOT.TH1"):
        py_h = TH1(h)
        neg_bin_indices = np.where(py_h.bin_content < 0.)
        neg_bin_values  = py_h.bin_content[neg_bin_indices]
        neg_bins = neg_bin_indices[0] + 1
        for i in neg_bins:
            # cast from np.int64 to int
            i = int(i)
            h.SetBinContent(i, 0)
            h.SetBinError(i, 0)
        return neg_bin_indices[0], neg_bin_values
    
    @staticmethod
    def GetBinContentArray(h, dtype:str='double', underflow_bin:int=0, overflow_bin:int=0):
        import ROOT
        c_vector = ROOT.TH1Utils.GetBinContentArray[dtype](h, underflow_bin, overflow_bin)
        return TArrayData.vec_to_array(c_vector)        
        
    @staticmethod
    def GetBinErrorArray(h, dtype:str='double', underflow_bin:int=0, overflow_bin:int=0):
        import ROOT
        c_vector = ROOT.TH1Utils.GetBinErrorArray[dtype](h, underflow_bin, overflow_bin)
        return TArrayData.vec_to_array(c_vector)
                     
    @staticmethod
    def GetBinErrorUpArray(h, dtype:str='double', underflow_bin:int=0, overflow_bin:int=0):
        import ROOT
        c_vector = ROOT.TH1Utils.GetBinErrorUpArray[dtype](h, underflow_bin, overflow_bin)
        return TArrayData.vec_to_array(c_vector)

    @staticmethod
    def GetBinErrorLowArray(h, dtype:str='double', underflow_bin:int=0, overflow_bin:int=0):
        import ROOT
        c_vector = ROOT.TH1Utils.GetBinErrorLowArray[dtype](h, underflow_bin, overflow_bin)
        return TArrayData.vec_to_array(c_vector)                     

    @staticmethod
    def GetBinCenterArray(h, dtype:str='double', underflow_bin:int=0, overflow_bin:int=0):
        import ROOT
        c_vector = ROOT.TH1Utils.GetBinCenterArray[dtype](h, underflow_bin, overflow_bin)
        return TArrayData.vec_to_array(c_vector)
    
    @staticmethod
    def GetBinWidthArray(h, dtype:str='double', underflow_bin:int=0, overflow_bin:int=0):
        import ROOT
        c_vector = ROOT.TH1Utils.GetBinWidthArray[dtype](h, underflow_bin, overflow_bin)
        return TArrayData.vec_to_array(c_vector)

    @staticmethod
    def GetBinLowEdgeArray(h, dtype:str='double', underflow_bin:int=0, overflow_bin:int=0):
        import ROOT
        c_vector = ROOT.TH1Utils.GetBinLowEdgeArray[dtype](h, underflow_bin, overflow_bin)
        return TArrayData.vec_to_array(c_vector)
    
    @staticmethod
    def GetPoissonError(data:np.ndarray, n_sigma:float=1):
        import ROOT
        data = np.array(data)
        c_data = as_vector(data)
        c_result  = ROOT.THistUtils.GetPoissonError(c_data, n_sigma)
        result = as_np_array(c_result)
        data_size = data.shape[0]
        pois_error = {
            'lo': result[:data_size],
            'hi': result[data_size:]
        }
        return pois_error
        
    @staticmethod
    def add_ghost_weights(h:"ROOT.TH1",
                          blind_range:Optional[List[float]]=None,
                          ghost_weight:float=1e-9):
        n_bins = h.GetNbinsX()
        for i in range(1, n_bins + 1):
            if h.GetBinContent(i) != 0:
                continue
            bin_center = h.GetBinCenter(i)
            if blind_range and (bin_center > blind_range[0]) and (bin_center < blind_range[1]):
                continue
            h.SetBinContent(i, ghost_weight)


def set_multithread(multithread:bool):
    import ROOT
    if (not ROOT.IsImplicitMTEnabled()) and (multithread):
        ROOT.EnableImplicitMT()
    elif (ROOT.IsImplicitMTEnabled()) and (not multithread):
        ROOT.DisableImplicitMT()

class RMultithreadEnv(object):
    def __init__(self, enable_multithread:bool):
        import ROOT
        self.original_multithread_state = ROOT.IsImplicitMTEnabled()
        self.new_multithread_state = enable_multithread
      
    def __enter__(self):
        import ROOT
        if (not self.original_multithread_state) and self.new_multithread_state:
            ROOT.EnableImplicitMT()
        elif (self.original_multithread_state) and (not self.new_multithread_state):
            ROOT.DisableImplicitMT()
        return self
  
    def __exit__(self, *args):
        import ROOT
        if (not self.original_multithread_state) and self.new_multithread_state:
            ROOT.DisableImplicitMT()
        elif (self.original_multithread_state) and (not self.new_multithread_state):
            ROOT.EnableImplicitMT()


from typing import Optional, Union, List, Dict

import ROOT

from quickstats import semistaticmethod, AbstractObject
from quickstats.utils.common_utils import combine_dict
from quickstats.utils.root_utils import load_macro

class ROOTObject(AbstractObject):   
    
    _DEFAULT_ROOT_CONFIG_ = {
        "SetBatch" : True,
        "TH1Sumw2" : None
    }
    
    _DEFAULT_ROOFIT_CONFIG_ = {
        "GlobalKillBelow"     : "ERROR",
        "MinimizerPrintLevel" : -1,
        "RemoveMessage"       : {
            "NumIntegration" : True,
            "Fitting"        : True,
            "Minimization"   : True,
            "InputArguments" : True,
            "Eval"           : True
        }
    }     
    
    _REQUIRE_CONFIG_ = {
        "ROOT"  : False,
        "RooFit": False
    }
    
    def __init__(self, root_config:Optional[Dict]=None,
                 roofit_config:Optional[Dict]=None,
                 verbosity:str="INFO"):
        super().__init__(verbosity=verbosity)
        self._root_config   = combine_dict(self._DEFAULT_ROOT_CONFIG_, root_config)
        self._roofit_config = combine_dict(self._DEFAULT_ROOFIT_CONFIG_, roofit_config)
        
        self._initialize()
    
    def configure_roofit(self):
        if "MinimizerPrintLevel" in  self._roofit_config:
            ROOT.Math.MinimizerOptions.SetDefaultPrintLevel(self._roofit_config["MinimizerPrintLevel"])
        if "RemoveMessage" in self._roofit_config:
            stream = ROOT.RooMsgService.instance().getStream(1)
            for topic, do_remove in self._roofit_config["RemoveMessage"].items():
                if not do_remove:
                    continue
                try:
                    roofit_topic = getattr(ROOT.RooFit, topic)
                    stream.removeTopic(roofit_topic)
                except:
                    self.stdout.warning(f'WARNING: Failed to remove RooFit message topic "{topic}".')
        if "GlobalKillBelow" in self._roofit_config:
            try:
                kill_level = self._roofit_config["GlobalKillBelow"]
                roofit_kill_level = getattr(ROOT.RooFit, kill_level)
                ROOT.RooMsgService.instance().setGlobalKillBelow(roofit_kill_level)
            except:
                self.stdout.warning(f'WARNING: Failed to set RooFit global kill below "{kill_level}"')
                
    def configure_root(self):
        set_batch = self._root_config.get("SetBatch", None)
        if set_batch is not None:
            ROOT.gROOT.SetBatch(set_batch)
        th1_sumw2 = self._root_config.get("TH1Sumw2", None)
        if th1_sumw2 is not None:
            ROOT.TH1.SetDefaultSumw2(th1_sumw2)
            
    def _initialize(self):
        if self._REQUIRE_CONFIG_["ROOT"]:
            self.configure_root()
        if self._REQUIRE_CONFIG_["RooFit"]:
            self.configure_roofit()
            
    @semistaticmethod
    def load_extension(self, name:str):
        result = load_macro(name)
        if (result is not None) and hasattr(ROOT, name):
            self.stdout.info(f'INFO: Loaded extension module "{name}"')            


from typing import Optional, Dict
import copy
from dataclasses import asdict, dataclass

class DataclassObject(object):
    
    @classmethod
    def from_dict(cls, source:Optional[Dict]=None):
        if source is None:
            return cls()
        return cls(**source)
    
    def to_dict(self):
        return asdict(self)
    
    def update(self, **kwargs):
        for key, value in kwargs.items():
            if hasattr(self, key):
                value_copy = copy.deepcopy(value)
                setattr(self, key, value_copy)


from .analysis_base import AnalysisBase
from .analysis_path_manager import AnalysisPathManager
from .ntuple_process_tool import NTupleProcessTool
from .ntuple_conversion_tool import NTupleConversionTool
from .multi_class_boundary_tree import MultiClassBoundaryTree
from .event_categorization import EventCategorization, CategorizationOutputMode, CategorizationOutputFormat


from typing import Optional, Union, Dict, List, Tuple, Callable
from cycler import cycler

import matplotlib

from quickstats.plots.color_schemes import QUICKSTATS_PALETTES
from quickstats.plots.template import (single_frame, parse_styles, format_axis_ticks,
                                       parse_analysis_label_options)
from quickstats.utils.common_utils import combine_dict, insert_periodic_substr

class AbstractPlot:
    
    STYLES = {}
    
    COLOR_CYCLE = QUICKSTATS_PALETTES['default']
    
    COLOR_PALLETE = {}
    COLOR_PALLETE_SEC = {}
    
    CONFIG = {
        "xlabellinebreak": 50,
        "ylabellinebreak": 50
    }
    
    def __init__(self,
                 color_pallete:Optional[Dict]=None,
                 color_pallete_sec:Optional[Dict]=None,
                 color_cycle:Optional[Dict]=None,
                 styles:Optional[Union[Dict, str]]=None,
                 analysis_label_options:Optional[Dict]=None,
                 figure_index:Optional[int]=None,
                 config:Optional[Dict]=None):
        
        self.color_pallete     = combine_dict(self.COLOR_PALLETE, color_pallete)
        self.color_pallete_sec = combine_dict(self.COLOR_PALLETE_SEC, color_pallete_sec)
            
        self.styles = combine_dict(self.STYLES, styles)
        self.styles = parse_styles(self.styles)
        
        if analysis_label_options is None:
            self.analysis_label_options = None
        else:
            self.analysis_label_options = parse_analysis_label_options(analysis_label_options)
            
        self.legend_order = self.get_default_legend_order()
        
        self.legend_data = {}
        self.legend_data_sec = {}
        self.legend_data_ext = {}
        
        self.figure_index = figure_index
        
        self.config = combine_dict(AbstractPlot.CONFIG, self.CONFIG)
        if config is not None:
            self.config = combine_dict(self.config, config)

        if color_cycle is None:
            self.prop_cycle = (cycler(color=self.COLOR_CYCLE))
        else:
            if isinstance(color_cycle, str):
                if color_cycle not in QUICKSTATS_PALETTES:
                    raise RuntimeError(f"unknown color palette: {color_cycle}")
                color_cycle = QUICKSTATS_PALETTES[color_cycle]
            self.prop_cycle = (cycler(color=color_cycle))
            
        self.annotation_list = []
        
    def add_annotation(self, text:str, **kwargs):
        self.annotation_list.append({"text": text, **kwargs})
        
    def get_default_legend_order(self):
        return []
    
    def update_legend_handles(self, handles:Dict, sec:bool=False,
                              idx:Optional[int]=None):
        if idx is None:
            if not sec:
                legend_data = self.legend_data
            else:
                legend_data = self.legend_data_sec
        else:
            if idx not in self.legend_data_ext:
                self.legend_data_ext[idx] = {}
            legend_data = self.legend_data_ext[idx]
            
        for key in handles:
            handle = handles[key]
            if isinstance(handle, matplotlib.container.Container):
                label = handle.get_label()
            elif isinstance(handle, (tuple, list)):
                label = handle[0].get_label()
            else:
                label = handle.get_label()               
            if label and not label.startswith('_'):
                legend_data[key] = {
                    'handle': handle,
                    'label': label
                }
            else:
                raise RuntimeError(f"the handle {handle} does not have an associated label")
                
    def add_legend_decoration(self, decorator, targets:List[str]):
        for key, legend_data in self.legend_data.items():
            if key not in targets:
                continue
            handle = legend_data["handle"]
            if isinstance(handle, tuple):
                new_handle = (*handle, decorator)
            else:
                new_handle = (handle, decorator)
            legend_data["handle"] = new_handle

    def get_legend_handles_labels(self, sec:bool=False,
                                  idx:Optional[Union[int, List[int]]]=None):
        handles = []
        labels = []
        if idx is None:
            if not sec:
                legend_data = self.legend_data
            else:
                legend_data = self.legend_data_sec        
            for key in self.legend_order:
                if key in legend_data:
                    handle = legend_data[key]['handle']
                    label = legend_data[key]['label']
                    handles.append(handle)
                    labels.append(label)
        else:
            if isinstance(idx, int):
                indices = [idx]
            else:
                indices = idx
            for index in indices:
                legend_data = self.legend_data_ext[index]
                for key in self.legend_order:
                    if key in legend_data:
                        handle = legend_data[key]['handle']
                        label = legend_data[key]['label']
                        handles.append(handle)
                        labels.append(label)
        return handles, labels
    
    def draw_frame(self, frame_method:Callable=None, **kwargs):
        if frame_method is None:
            frame_method = single_frame
        ax = frame_method(styles=self.styles,
                          prop_cycle=self.prop_cycle,
                          analysis_label_options=self.analysis_label_options,
                          figure_index=self.figure_index,
                          **kwargs)
        for annotation_kwargs in self.annotation_list:
            annotation_kwargs = combine_dict(self.styles['annotation'], annotation_kwargs)
            if isinstance(ax, tuple):
                ax[0].annotate(**annotation_kwargs)
            else:
                ax.annotate(**annotation_kwargs)
        self.figure = matplotlib.pyplot.gcf()
        return ax
    
    def draw_axis_labels(self, ax, xlabel:Optional[str]=None, ylabel:Optional[str]=None,
                         xlabellinebreak:Optional[int]=None, ylabellinebreak:Optional[int]=None):
        if xlabel is not None:
            if (xlabellinebreak is not None) and (xlabel.count("$") < 2):
                xlabel = insert_periodic_substr(xlabel, xlabellinebreak)
            ax.set_xlabel(xlabel, **self.styles['xlabel'])
        if ylabel is not None:
            if (ylabellinebreak is not None) and (ylabel.count("$") < 2):
                ylabel = insert_periodic_substr(ylabel, ylabellinebreak)            
            ax.set_ylabel(ylabel, **self.styles['ylabel'])
            
    def draw_axis_components(self, ax, xlabel:Optional[str]=None, ylabel:Optional[str]=None,
                             ylim:Optional[Tuple[float]]=None, xlim:Optional[Tuple[float]]=None):
        
        self.draw_axis_labels(ax, xlabel, ylabel,
                              xlabellinebreak=self.config["xlabellinebreak"],
                              ylabellinebreak=self.config["ylabellinebreak"],)
        
        format_axis_ticks(ax, **self.styles['axis'],
                          xtick_styles=self.styles['xtick'],
                          ytick_styles=self.styles['ytick'])
        
        if ylim is not None:
            ax.set_ylim(*ylim)
        if xlim is not None:
            ax.set_xlim(*xlim)
    
    def set_axis_range(self, ax,
                       xmin:Optional[float]=None, xmax:Optional[float]=None,
                       ymin:Optional[float]=None, ymax:Optional[float]=None,
                       ypad:Optional[float]=None):
        xlim = list(ax.get_xlim())
        ylim = list(ax.get_ylim())
        if xmin is not None:
            xlim[0] = xmin
        if xmax is not None:
            xlim[1] = xmax
        if ypad is not None:
            if (ymax is not None):
                raise ValueError('can not use both "ypad" and "ymax" when setting axis range')
            if ypad < 0 or ypad > 1:
                raise ValueError('"ypad" must be between 0 and 1')
            ylim[1] = ylim[1] / (1 - ypad)
        if ymin is not None:
            ylim[0] = ymin
        if ymax is not None:
            ylim[1] = ymax                
        ax.set_xlim(*xlim)
        ax.set_ylim(*ylim)


from typing import Optional, List
import fnmatch

from .rooproc_hybrid_action import RooProcHybridAction

from quickstats.utils.common_utils import is_valid_file

class RooProcSave(RooProcHybridAction):
    
    def __init__(self, treename:str, filename:str, 
                 columns:Optional[List[str]]=None,
                 frame:Optional[str]=None):
        super().__init__(treename=treename,
                         filename=filename,
                         columns=columns,
                         frame=frame)
        
    @classmethod
    def parse(cls, main_text:str, block_text:Optional[str]=None):
        kwargs = cls.parse_as_kwargs(main_text)
        return cls(**kwargs)
    
    def _execute(self, rdf:"ROOT.RDataFrame", processor:"quickstats.RooProcessor", **params):
        treename = params['treename']
        filename = params['filename']
        if processor.cache and is_valid_file(filename):
            processor.stdout.info(f'INFO: Cached output from "{filename}".')
            return rdf, processor
        columns = params.get('columns', None)
        if isinstance(columns, str):
            columns = self.parse_as_list(columns)
        if columns is None:
            if processor.use_template:
                from quickstats.utils.root_utils import templated_rdf_snapshot
                rdf_next = templated_rdf_snapshot(rdf)(treename, filename)
            else:
                rdf_next = rdf.Snapshot(treename, filename)
        else:
            all_columns = [str(c) for c in rdf.GetColumnNames()]
            save_columns = []
            for column in columns:
                save_columns += [c for c in all_columns if fnmatch.fnmatch(c, column)]
            save_columns = list(set(save_columns))
            self.makedirs(filename)
            if processor.use_template:
                from quickstats.utils.root_utils import templated_rdf_snapshot 
                rdf_next = templated_rdf_snapshot(rdf, save_columns)(treename, filename, save_columns)
            else:
                rdf_next = rdf.Snapshot(treename, filename, save_columns)
        processor.stdout.info(f'INFO: Writing output to "{filename}".')
        return rdf_next, processor


from typing import Optional, Union, List, Dict
import builtins

import pandas as pd
import numpy as np

from quickstats.plots import AbstractPlot

class Histo1DPlot(AbstractPlot):
    
    STYLES = {
        'hist':{
            'histtype': 'stepfilled',
            'rwidth': 0,
            'linewidth': 2,
            'edgecolor': 'k'
        }    
    } 
    
    def __init__(self, data:Union[pd.DataFrame, Dict[str, np.ndarray], np.ndarray, List],
                 label_map:Optional[Dict[str, str]]=None,
                 color_cycle:Optional[Union[Dict, str]]="hist_series",
                 styles:Optional[Union[Dict, str]]=None,
                 analysis_label_options:Optional[Dict]=None,
                 config:Optional[Dict]=None):
        
        if not isinstance(data, (list, np.ndarray, pd.DataFrame, dict)):
            raise ValueError(f"unsupported data type: {type(data)}")
        self.data = data
        self.label_map = label_map
        
        super().__init__(color_cycle=color_cycle,
                         styles=styles,
                         analysis_label_options=analysis_label_options,
                         config=config)
        
    def draw(self, columns:Optional[Union[str, List[str]]]=None,
             scale:float=None, bins:int=None, range=None, stacked:bool=False,
             density:bool=False, ypad:Optional[float]=None,
             xlabel:Optional[str]=None, ylabel:Optional[str]="Events",
             xmin:Optional[float]=None, xmax:Optional[float]=None,
             logy:bool=False):
        
        if isinstance(columns, str):
            columns = [columns]
        
        x_collection = []
        if isinstance(self.data, (list, np.ndarray)):
            if columns is not None:
                raise ValueError("specification of column(s) are not allowed with single data array as input")
            x_collection.append(np.array(self.data))
        elif isinstance(self.data, pd.DataFrame):
            for column in columns:
                if column not in self.data.columns:
                    raise RuntimeError(f"dataframe has no column named \"{column}\"")
                x = self.data[column].values
                x = x[np.isfinite(x)]
                x_collection.append(x)
        elif isinstance(self.data, dict):
            for column in columns:
                if column not in self.data:
                    raise RuntimeError(f"data dictionary has no key named \"{column}\"")
                x = self.data[column]
                x = x[np.isfinite(x)]
                x_collection.append(x)
        else:
            raise ValueError(f"unsupported data type: {type(self.data)}")
            
        if scale is not None:
            for i in builtins.range(len(x_collection)):
                x_collection[i] *= scale
                
        n_columns = len(x_collection)
        if n_columns > 1:
            if self.label_map is None:
                labels = list(columns)
            else:
                labels = [self.label_map.get(column, column) for column in columns]
        else:
            labels = None
        
        ax = self.draw_frame(logy=logy)
        
        n, bins, patches = ax.hist(x_collection, bins=bins, range=range,
                                   stacked=stacked, density=density,
                                   label=labels, **self.styles["hist"])
        
        if n_columns > 1:
            ax.legend(**self.styles['legend'])
        # update axis offset text
        import matplotlib.pyplot as plt
        plt.tight_layout()
        self.draw_axis_components(ax, xlabel=xlabel, ylabel=ylabel)
        self.set_axis_range(ax, xmin=xmin, xmax=xmax, ypad=ypad)  
        
        return ax


from typing import Optional, Union, Dict, List

import pandas as pd
import numpy as np

from matplotlib import colors
from matplotlib.ticker import MaxNLocator
from matplotlib.lines import Line2D
from matplotlib.patches import Polygon

from quickstats.plots import AbstractPlot
from quickstats.plots.template import single_frame, parse_styles, create_transform, format_axis_ticks
from quickstats.utils.common_utils import combine_dict


class ScoreDistributionPlot(AbstractPlot):
    
    CONFIG = {
        "boundary_style": {
            "ymin": 0,
            "ymax": 0.5,
            "linestyle": "--",
            "color": "k"
        }
    }
    
    def __init__(self, data_map:Dict[str, pd.DataFrame], plot_options:Dict[str, Dict],
                 styles:Optional[Union[Dict, str]]=None,
                 analysis_label_options:Optional[Dict]=None,
                 config:Optional[Dict]=None):
        """
        Arguments:
            plot_options: dicionary
                A dictionary containing plot options for various group of samples.
                Format: { <sample_group>: {
                            "samples": <list of sample names>,
                            "style": <options in mpl.hist>},
                            "type": "hist" or "errorbar"
                          ...}
             
        """
        self.data_map = data_map
        self.plot_options = plot_options
        super().__init__(styles=styles,
                         analysis_label_options=analysis_label_options,
                         config=config)
    
    def draw(self, column_name:str="score", weight_name:Optional[str]="weight",
             xlabel:str="Score", ylabel:str="Fraction of Events / {bin_width:.2f}",
             boundaries:Optional[List]=None, nbins:int=25, xmin:float=0, xmax:float=1):
        """
        
        Arguments:
            column_name: string, default = "score"
                Name of the score variable in the dataframe.
            weight_name: (optional) string, default = "weight"
                If specified, normalize the histogram by the "weight_name" variable
                in the dataframe.
            xlabel: string, default = "Score"
                Label of x-axis.
            ylabel: string, default = "Fraction of Events / {bin_width}"
                Label of y-axis.
            boundaries: (optional) list of float
                If specified, draw score boundaries at given values.
            nbins: int, default = 25
                Number of histogram bins.
            xmin: float, default = 0
                Minimum value of x-axis.
            xmax: float, default = 1
                Maximum value of x-axis.
        """
        ax = self.draw_frame()
        for key in self.plot_options:
            samples = self.plot_options[key]["samples"]
            plot_style  = self.plot_options[key].get("style", {})
            df = pd.concat([self.data_map[sample] for sample in samples], ignore_index = True)
            if weight_name is not None:
                norm_weights = df[weight_name] / df[weight_name].sum()
            else:
                norm_weights = None
            plot_type = self.plot_options[key].get("type", "hist")
            if plot_type == "hist":
                y, x, _ = ax.hist(df[column_name], nbins, range=(xmin, xmax),
                                  weights=norm_weights, **plot_style, zorder=-5)
            elif plot_type == "errorbar":
                n_data = len(df[column_name])
                norm_weights = np.ones((n_data,)) / n_data
                y, bins = np.histogram(df[column_name], nbins, weights=norm_weights)
                bin_centers  = 0.5*(bins[1:] + bins[:-1])
                from quickstats.maths.statistics import get_poisson_interval
                yerr = get_poisson_interval(y * n_data)
                ax.errorbar(bin_centers, y, 
                            yerr=(yerr["lo"] / n_data, yerr["hi"] / n_data),
                            **plot_style)
            else:
                raise RuntimeError(f'unknown plot type: {plot_type}')
                
        bin_width = (xmax - xmin) / nbins
        ylabel = ylabel.format(bin_width=bin_width)
        
        self.draw_axis_components(ax, xlabel=xlabel, ylabel=ylabel)
        self.set_axis_range(ax, xmin=xmin, xmax=xmax)
        
        ax.yaxis.set_major_locator(MaxNLocator(prune='lower', steps=[10]))
        ax.xaxis.set_major_locator(MaxNLocator(steps=[10]))

        handles, labels = ax.get_legend_handles_labels()
        new_handles = [Line2D([], [], c=h.get_edgecolor(), linestyle=h.get_linestyle(),
                              **self.styles['legend_Line2D']) if isinstance(h, Polygon) else h for h in handles]
        ax.legend(handles=new_handles, labels=labels, **self.styles['legend'])
        if boundaries is not None:
            for boundary in boundaries:
                ax.axvline(x=boundary, **self.config["boundary_style"])
        return ax


def score_distribution_plot(dfs:Dict[str, pd.DataFrame], hist_options:Dict[str, Dict], 
                            data_options:Optional[Dict[str, Dict]]=None,
                            nbins:int=25, xmin:float=0, xmax:float=1, score_name:str='score', weight_name:str='weight',
                            xlabel:str='NN Score', ylabel:str='Fraction of Events / {bin_width}',
                            boundaries:Optional[List]=None, plot_styles:Optional[Dict]=None,
                            analysis_label_options:Optional[Dict]=None):
    styles = parse_styles(plot_styles)
    ax = single_frame(styles=styles, analysis_label_options=analysis_label_options)
    for key in hist_options:
        samples = hist_options[key]['samples']
        hist_style     = hist_options[key].get('style', {})
        combined_df = pd.concat([dfs[sample] for sample in samples], ignore_index = True)
        norm_weights = combined_df[weight_name]/combined_df[weight_name].sum()
        y, x, _ = ax.hist(combined_df[score_name], nbins, range=(xmin, xmax), weights=norm_weights, **hist_style,
                          zorder=-5)
    if data_options is not None:
        for key in data_options:
            samples = data_options[key]['samples']
            errorbar_style     = data_options[key].get('style', {})
            combined_df = pd.concat([dfs[sample] for sample in samples], ignore_index = True)
            norm_weights = combined_df[weight_name]/combined_df[weight_name].sum()
            y, bins = np.histogram(combined_df[score_name], nbins, weights=norm_weights)
            bin_centers  = 0.5*(bins[1:] + bins[:-1])
            ax.errorbar(bin_centers, y, yerr=y**0.5, **errorbar_style)
    ax.yaxis.set_major_locator(MaxNLocator(prune='lower', steps=[10]))
    ax.xaxis.set_major_locator(MaxNLocator(steps=[10]))
    ax.set_xlim(xmin, xmax)
    bin_width = 1/nbins
    ax.set_xlabel(xlabel, **styles['xlabel'])
    ax.set_ylabel(ylabel.format(bin_width=bin_width), **styles['ylabel'])
    handles, labels = ax.get_legend_handles_labels()
    new_handles = [Line2D([], [], c=h.get_edgecolor(), linestyle=h.get_linestyle(), **styles['legend_Line2D'])
                   if isinstance(h, Polygon) else h for h in handles]
    ax.legend(handles=new_handles, labels=labels, **styles['legend'])
    if boundaries is not None:
        for boundary in boundaries:
            ax.axvline(x=boundary, ymin=0, ymax=0.5, linestyle='--', color='k')
    return ax

def score_distribution_plot_2D(dfs:Dict[str, pd.DataFrame], hist_options:Dict[str, Dict], 
                            data_options:Optional[Dict[str, Dict]]=None,
                            xbins:int=25, xmin:float=0, xmax:float=1, ybins:int=25, ymin:float=0,
                            ymax:float=1, xscore_name:str='score', yscore_name:str='score', weight_name:str='weight',
                            xlabel:str='NN Score 1', ylabel:str='NN Score 2', zlabel:str='Fraction of Events / bin',
                            coordinates:Optional[List]=None, plot_styles:Optional[Dict]=None,
                            analysis_label_options:Optional[Dict]=None, show_text=False, density=False):
    styles = parse_styles(plot_styles)
    ax = single_frame(styles=styles, analysis_label_options=analysis_label_options)
    key = list(hist_options.keys())[0]
    samples = hist_options[key]['samples']
    hist_style = hist_options[key].get('style_2D', {})
    if hist_style.pop('logz', False):
        log_scale = colors.LogNorm()
    else:
        log_scale = colors.Normalize()
    combined_df = pd.concat([dfs[sample] for sample in samples], ignore_index = True)
    norm_weights = combined_df[weight_name]
    if density:
        norm_weights /= norm_weights.sum()
    h, x, y, image = ax.hist2d(x=combined_df[xscore_name], y=combined_df[yscore_name], bins=[xbins, ybins],
                        range=[[xmin, xmax], [ymin, ymax]], weights=norm_weights, **hist_style, 
                        norm=log_scale, zorder=-5)
    if show_text:
        for i in range(len(y)-1):
            for j in range(len(x)-1):
                ax.text((x[j]+x[j+1])/2,(y[i]+y[i+1])/2, h.T[i,j], color="r", ha="center", va="center", fontweight="bold")

    if coordinates is not None:
        for coordinate in coordinates:
            ax.plot([i[0] for i in coordinate], [i[1] for i in coordinate], 'r--')
    ax.xaxis.set_major_locator(MaxNLocator(steps=[10]))
    ax.yaxis.set_major_locator(MaxNLocator(steps=[10]))
    ax.set_xlim(xmin, xmax)
    ax.set_ylim(ymin, ymax)
    ax.set_xlabel(xlabel, **styles['xlabel'])
    ax.set_ylabel(ylabel, **styles['ylabel'])
    import matplotlib.pyplot as plt
    cbar = plt.colorbar(image, ax=ax, pad=0.02)
    format_axis_ticks(cbar.ax, x_axis=False, y_axis=True, ytick_styles=styles['ytick'], **styles['z-axis'])
    cbar.ax.yaxis.set_label_position("right")
    cbar.ax.yaxis.tick_right()
    cbar.ax.set_ylabel(zlabel, **styles['zlabel'])
    return ax



from typing import Optional, List, Dict, Union, Callable
import os
import glob
import json

import numpy as np

from quickstats import ConfigurableObject, semistaticmethod
from .analysis_path_manager import AnalysisPathManager
from .config_format_templates import DEFAULT_SAMPLE_CONFIG_FORMAT

class NTupleProcessTool(ConfigurableObject):
    """ Tool for processing root ntuples
    """
    
    CONFIG_FORMAT = DEFAULT_SAMPLE_CONFIG_FORMAT
    
    REQUIRED_CONFIG_COMPONENTS = ["sample_dir", "sample_subdir", "samples", "merge_samples"]
    
    DEFAULT_SAMPLE_OUTNAME         = "{sample_name}_{sample_type}.root"
    DEFAULT_MERGED_SAMPLE_OUTNAME  = "{sample_name}.root"
    DEFAULT_CUTFLOW_OUTNAME        = "cutflow_{sample_name}_{sample_type}.csv"
    DEFAULT_WEIGHT_OUTNAME         = "yield_{sample_name}_{sample_type}.json"
    DEFAULT_MERGED_CUTFLOW_OUTNAME = "cutflow_{sample_name}.csv"
    
    @property
    def sample_config(self):
        return self.config
    
    def __init__(self, sample_config:Union[Dict, str], outdir:str='output',
                 processor_config:Optional[str]=None,
                 processor_flags:Optional[List[str]]=None,
                 use_template:bool=False,
                 multithread:bool=True,
                 disable_config_message:bool=False,
                 verbosity:Optional[Union[int, str]]="INFO",
                 **kwargs):
        
        super().__init__(disable_config_message=disable_config_message,
                         verbosity=verbosity)
        self.outdir = outdir
        self.path_manager = AnalysisPathManager(outdir)
        self.path_manager.set_directory("ntuple", "ntuples")
        self.path_manager.set_directory("cutflow", "cutflow")
        
        self.load_sample_config(sample_config)
        
        self.processor = None        
        if processor_config is not None:
            self.load_processor_config(processor_config,
                                       use_template=use_template,
                                       multithread=multithread)
            
        if processor_flags is not None:
            self.processor_flags = list(processor_flags)
        else:
            self.processor_flags = []
        
        self.cutflow_report = None
        
    def load_sample_config(self, config_source:Union[Dict, str]):
        if isinstance(config_source, str):
            if not os.path.exists(config_source):
                raise FileNotFoundError(f'config file "{config_source}" does not exist')
            config_path = os.path.abspath(config_source)
            self.path_manager.set_file("sample_config", config_path)
        self.load_config(config_source)
        
        self.all_samples = list(self.sample_config['samples'])
        merge_samples = self.sample_config['merge_samples']
        if not merge_samples:
            self.all_merged_samples = list(self.all_samples)
        else:
            self.all_merged_samples = list(merge_samples)
            # fill missing samples
            for sample in self.all_samples:
                if ((sample not in merge_samples) and 
                    all(sample not in _samples for _samples in merge_samples.values())):
                    self.all_merged_samples.append(sample)
        
    def load_processor_config(self, config_path:str,
                              multithread:bool=True,
                              use_template:bool=False):
        from quickstats.components.processors import RooProcessor
        self.processor = RooProcessor(config_path,
                                      use_template=use_template,
                                      multithread=multithread,
                                      verbosity=self.stdout.verbosity)
        self.path_manager.set_file("processor_config", os.path.abspath(config_path))
        
    def get_validated_samples(self, samples:Optional[List[str]]=None, merged:bool=False):
        if merged:
            ref_samples = self.all_merged_samples
        else:
            ref_samples = self.all_samples
            
        if samples is not None:
            for sample in samples:
                if sample not in ref_samples:
                    raise RuntimeError(f'no information regarding the sample "{sample}" '
                                       f'is specified in the sample config file')
        else:
            samples = ref_samples
        return samples
    
    @semistaticmethod
    def get_sample_outpath(self, sample_config:Dict, outdir:str):
        sample_name = sample_config["name"]
        sample_type = sample_config["type"]
        basename = self.DEFAULT_SAMPLE_OUTNAME.format(sample_name=sample_name,
                                                      sample_type=sample_type)
        return os.path.join(outdir, basename)
    
    @semistaticmethod
    def get_merged_sample_outpath(self, sample_config:Dict, outdir:str):
        sample_name = sample_config["name"]
        basename = self.DEFAULT_MERGED_SAMPLE_OUTNAME.format(sample_name=sample_name)        
        return os.path.join(outdir, basename)
    
    @semistaticmethod
    def get_cutflow_outpath(self, sample_config:Dict, outdir:str):
        sample_name = sample_config["name"]
        sample_type = sample_config["type"]
        basename_cutflow = self.DEFAULT_CUTFLOW_OUTNAME.format(sample_name=sample_name,
                                                               sample_type=sample_type)
        basename_weight  = self.DEFAULT_WEIGHT_OUTNAME.format(sample_name=sample_name,
                                                              sample_type=sample_type)
        return [os.path.join(outdir, basename_cutflow), os.path.join(outdir, basename_weight)]  
    
    @semistaticmethod
    def get_merged_cutflow_outpath(self, sample_config:Dict, outdir:str):
        sample_name = sample_config["name"]
        basename = self.DEFAULT_MERGED_CUTFLOW_OUTNAME.format(sample_name=sample_name)
        return os.path.join(outdir, basename)    
    
    def prerun_process(self, sample_config:Dict):
        pass

    def process_samples(self, samples:Optional[List[str]]=None,
                        sample_types:Optional[List[str]]=None):
        if self.processor is None:
            raise RuntimeError("processor not initialized (probably missing a processor config)")
        samples       = self.get_validated_samples(samples)
        sample_dir    = self.sample_config.get('sample_dir', './')
        sample_subdir = self.sample_config.get('sample_subdir', {})
        sample_specs  = self.sample_config['samples']
        outdir = self.path_manager.base_path
        for sample in samples:
            self.processor.set_flags(self.processor_flags)
            for sample_type in sample_specs[sample]:
                if (sample_types is not None) and (sample_type not in sample_types):
                    continue
                all_files = []
                subsamples = sample_specs[sample][sample_type]
                if isinstance(subsamples, str):
                    subsamples = [subsamples]
                for subsample in subsamples:
                    sample_path = os.path.join(sample_dir, sample_subdir.get(sample_type, ''), 
                                               subsample)
                    if os.path.isdir(sample_path):
                        subfiles = glob.glob(os.path.join(sample_path, "*.root"))
                    else:
                        subfiles = glob.glob(sample_path)
                    if len(subfiles) == 0:
                        raise RuntimeError(f"no matching samples found for {sample_path}")
                    all_files += subfiles
                sample_config = {
                    "name": sample,
                    "path": sample_path,
                    "type": sample_type
                }
                self.processor.global_variables['sample'] = sample
                self.processor.global_variables['sample_type'] = sample_type
                self.processor.global_variables['outdir'] = outdir
                self.prerun_process(sample_config)                
                self.processor.run(all_files)
                self.processor.clear_global_variables()
                
    def merge_outputs(self, source_path_func:Callable,
                      target_path_func:Callable,
                      merge_func:Callable,
                      outdir:str,
                      samples:Optional[List[str]]=None,
                      subdirs:Optional[List[str]]=None):
        samples = self.get_validated_samples(samples, merged=True)
        sample_specs = self.sample_config['samples']
        if 'merge_samples' in self.sample_config:
            merged_sample_specs = self.sample_config['merge_samples']
        else:
            merged_sample_specs = {sample: [sample] for sample in samples}
        for sample in samples:
            if (sample not in merged_sample_specs) and \
               all(sample not in _samples for _samples in merged_sample_specs.values()):
                merged_sample_specs[sample] = [sample]
        if subdirs is None:
            subdirs = [""]
        for subdir in subdirs:
            sample_subdir = os.path.join(outdir, subdir)
            for sample in merged_sample_specs:
                # only merge what is requested
                if sample not in samples:
                    continue
                subsamples = merged_sample_specs[sample]
                source_files = []
                for subsample in subsamples:
                    # merge n-tuples from different mc campaigns / data years
                    for sample_type in sample_specs[subsample]:
                        sample_config = {
                            "name": subsample,
                            "type": sample_type
                        }
                        file = source_path_func(sample_config, sample_subdir)
                        source_files.append(file)
                sample_config = {"name": sample}
                target_file = target_path_func(sample_config, sample_subdir)
                self.stdout.info(f'INFO: Merging outputs for the sample "{sample}"')
                merge_func(source_files, target_file)                      
    
    def merge_samples(self, samples:Optional[List[str]]=None, subdirs:Optional[List[str]]=None):
        outdir = self.path_manager.get_directory("ntuple")
        self.merge_outputs(self.get_sample_outpath,
                           self.get_merged_sample_outpath,
                           self._merge_samples_with_hadd,
                           outdir=outdir,
                           samples=samples,
                           subdirs=subdirs)
    
    def _merge_samples_with_hadd(self, fnames:List[str], outname:str):
        for fname in fnames:
            if not os.path.exists(fname):
                raise FileNotFoundError(f'missing ntuple file "{fname}"')
        hadd_cmd = "hadd -f {} {}".format(outname, " ".join(fnames))
        os.system(hadd_cmd)
            
    def merge_cutflows(self, samples:Optional[List[str]]=None, subdirs:Optional[List[str]]=None):
        outdir = self.path_manager.get_directory("cutflow")
        self.merge_outputs(self.get_cutflow_outpath,
                           self.get_merged_cutflow_outpath,
                           self._merge_cutflow_data,
                           outdir=outdir,
                           samples=samples,
                           subdirs=subdirs)
            
    def _merge_cutflow_data(self, fnames:List[str], outname:str):
        import pandas as pd
        merged_df = None
        for fname in fnames:
            if isinstance(fname, (list, tuple)):
                # first file contains cutflow information
                # second file contains weight information                
                assert len(fname) == 2
                cutflow_fname = fname[0]
                weight_fname = fname[1]
            else:
                cutflow_fname = fname
                weight_fname = None
            if not os.path.exists(cutflow_fname):
                raise FileNotFoundError(f'missing cutflow file "{cutflow_fname}"')
            cutflow_df = pd.read_csv(cutflow_fname).set_index(['name'])
            if (weight_fname is None) or not os.path.exists(weight_fname):
                weight_data = None
            else:
                with open(weight_fname, 'r') as file:
                    weight_data = np.array(list(json.load(file).values()))
            if merged_df is None:
                merged_df = cutflow_df.copy()
                if weight_data is not None:
                    assert len(merged_df) == len(weight_data)
                    merged_df['yield'] = weight_data
            else:
                merged_df['all']  += cutflow_df['all']
                merged_df['pass'] += cutflow_df['pass']
                if 'yield' in merged_df:
                    if weight_data is None:
                        raise RuntimeError(f'missing weight file that is in association with the '
                                           f'cutflow file "{cutflow_fname}"')
                    merged_df['yield'] += weight_data
        if 'yield' in merged_df:
            yields = merged_df['yield'].values
            yield_efficiency = [100]
            yield_cumul_efficiency = [100]
            for i in range(1, len(yields)):
                yield_efficiency.append(100 * (yields[i] / yields[i-1]))
                yield_cumul_efficiency.append(100 * (yields[i] / yields[0]))
            merged_df['yield_efficiency'] = yield_efficiency
            merged_df['yield_cumulative_efficiency'] = yield_cumul_efficiency
        merged_df.to_csv(outname)
        self.stdout.info(f'INFO: Saved cutflow data as "{outname}"')
        
    def load_cutflow_report(self, samples:Optional[List[str]]=None):
        import pandas as pd
        samples = self.get_validated_samples(samples, merged=True)
        
        sample_spec   = self.sample_config['samples']
        cutflow_dir   = self.path_manager.get_directory("cutflow")
        
        cutflow_report = {}
        for sample in self.all_merged_samples:
            if sample not in samples:
                continue
            sample_config = {
                "name": sample
            }
            cutflow_file = self.get_merged_cutflow_outpath(sample_config, cutflow_dir)
            if not os.path.exists(cutflow_file):
                self.stdout.warning(f'WARNING: Missing cutflow file for the sample "{sample}".')
                continue
            df = pd.read_csv(cutflow_file)
            cutflow_report[sample] = df
            
        self.cutflow_report = cutflow_report

    def plot_cutflow_report(self, samples:Optional[List[str]]=None,
                            label_map:Optional[Dict]=None):

        import matplotlib.pyplot as plt
        
        plt.rcParams['figure.dpi'] = 200
        
        # Load the cutflow report if it doesn't exist yet
        if self.cutflow_report is None:
            self.load_cutflow_report(samples)

        colors = ["#36B1BF","#F2385A","#FDC536"]  # turqouise, pink, yellow

        if samples is None:
            samples = list(self.cutflow_report.keys())
            
        if label_map is None:
            label_map = {}
            
        # Loop over each sample and make a plot of the cutflow
        for sample in samples: 
            
            if sample not in self.cutflow_report:
                print(f'WARNING: Missing cutflow report for the sample "{sample}". Skipped.')
                continue

            df = self.cutflow_report[sample] 

            cut_labels = [label_map.get(name, name) for name in df["name"]]

            fig, axs = plt.subplots(3, 1, figsize=(17, 8))
            fig.tight_layout(pad=3.0)

            for i in range(len(axs)): # set x-axes style
                axs[i].tick_params(axis='x', rotation =15, labelsize=10) 

            # Yields bar chart 
            axs[0].bar(cut_labels, df["yield"], color=colors[0])
            axs[0].set_ylabel("Yields", fontsize=14)
            
            # Efficiency bar chart 
            axs[1].bar(cut_labels, df["yield_efficiency"], color=colors[1])
            axs[1].set_ylabel("Efficiency [%]", fontsize=14)
            
            # Culmulative efficiency bar chart 
            axs[2].bar(cut_labels, df["yield_cumulative_efficiency"], color=colors[2])
            axs[2].set_ylabel("Cumulative Efficiency [%]", fontsize=14)

            fig.suptitle(f"Sample: {sample}", fontsize=20, y=0.98)
            
            plt.show()


import numpy as np

def get_intersections(x:np.ndarray, y:np.ndarray, level:float, delta:float=0.0001):
    from scipy.interpolate import interp1d
    func_interp = interp1d(x, y)
    x_interp = np.arange(min(x), max(x), delta)
    y_interp = func_interp(x_interp)

    asign = np.sign(y_interp - level)
    sign_change   = (np.roll(asign, 1) - asign) != 0
    intersections = x_interp[sign_change]
    sign_slope    = asign[sign_change]
    # no intersections
    if len(intersections) == 0:
        return []
    if len(intersections) == 1:
        if sign_slope[0] == -1:
            return np.array([[intersections[0], np.inf]])
        else:
            return np.array([[-np.inf, intersections[0]]])
    else:
        if sign_slope[0] == 1:
            intersections = np.insert(intersections, 0, -np.inf)
        if sign_slope[-1] == -1:
            intersections = np.insert(intersections, intersections.shape[0], np.inf)
        if len(intersections) & 1:
            raise RuntimeError("number of intersections can not be odd")
        n_pairs = len(intersections) // 2
        return intersections.reshape((n_pairs, 2))